import json
import logging
import time
from datetime import datetime, timedelta
from functools import cache
from pathlib import Path
from typing import Any, Optional

import requests

from atopile import address, errors, instance_methods, config

from atopile.address import AddrStr
from atopile.front_end import RangedValue

log = logging.getLogger(__name__)


def _get_specd_mpn(addr: AddrStr) -> str:
    """
    Return the MPN for a component given its address
    """
    comp_data = instance_methods.get_data_dict(addr)

    try:
        return comp_data["mpn"]
    except KeyError as ex:
        raise MissingData("$addr has no MPN", title="No MPN", addr=addr) from ex


def _get_specd_type(addr: AddrStr) -> str:
    """
    Return the type for a component given its address
    """
    try:
        mpn = _get_specd_mpn(addr)
        # split off "generic_" from the mpn
        return mpn.split("_", 1)[1]
    except KeyError as ex:
        raise MissingData("$addr has no type", title="No Type", addr=addr) from ex


def _is_generic(addr: AddrStr) -> bool:
    """
    Return whether a component is generic
    """
    # check if "generic_" is in the mpn
    return _get_specd_mpn(addr).startswith("generic_")


class NoMatchingComponent(errors.AtoError):
    """
    Raised when there's no component matching the given parameters in jlc_parts.csv
    """

    title = "No component matches parameters"


component_cache: dict[str, Any]
cache_file_path: Path

def configure_cache(top_level_path: Path):
    """Configure the cache to be used by the component module."""
    global component_cache
    global cache_file_path
    cache_file_path = top_level_path / ".ato/component_cache.json"
    if cache_file_path.exists():
        with open(cache_file_path, "r") as cache_file:
                component_cache = json.load(cache_file)
        # Clean out stale entries
        clean_cache()
    else:
        component_cache = {}

def save_cache():
    """Saves the current state of the cache to a file."""
    cache_file_path.parent.mkdir(parents=True, exist_ok=True)
    with open(cache_file_path, "w") as cache_file:
        # Convert the ChainMap to a regular dictionary
        serializable_cache = dict(component_cache)
        json.dump(serializable_cache, cache_file)


def get_component_from_cache(component_addr: AddrStr, current_data: dict) -> Optional[dict]:
    """Retrieve a component from the cache, if available, not stale, and unchanged."""
    if component_addr not in component_cache:
        return None

    # Check the cache age
    cached_entry = component_cache[component_addr]
    cached_timestamp = datetime.fromtimestamp(cached_entry["timestamp"])
    cache_age = datetime.now() - cached_timestamp
    if cache_age > timedelta(days=14):
        return None

    # Check the component attrs
    if current_data != cached_entry["address_data"]:
        return None

    log.debug("Using cached component for %s", component_addr)
    return cached_entry["data"]


def update_cache(component_addr, component_data, address_data):
    """Update the cache with new component data and save it."""
    component_cache[component_addr] = {
        "data": component_data,
        "timestamp": time.time(),  # Current time as a timestamp
        "address_data": dict(address_data),  # Source attributes used to detect changes
    }
    save_cache()


def clean_cache():
    """Clean out entries older than 1 day."""
    addrs_to_delete = set()
    for addr, entry in component_cache.items():
        cached_timestamp = datetime.fromtimestamp(entry["timestamp"])
        if datetime.now() - cached_timestamp >= timedelta(days=1):
            addrs_to_delete.add(addr)

    for addr in addrs_to_delete:
        del component_cache[addr]

    save_cache()

@cache
def _get_generic_from_db(component_addr: str) -> dict[str, Any]:
    """
    Return the MPN for a component given its address
    """
    log.debug("Fetching component for %s", component_addr)

    specd_data = instance_methods.get_data_dict(component_addr)

    specd_data_dict = {
        k: v.to_dict() if isinstance(v, RangedValue) else v for k, v in specd_data.items()
    }

    # check if there are any Physical objects in the specd_data, if not, throw a warning
    if specd_data:  # Check that specd_data is not empty
        if not any(isinstance(v, RangedValue) for v in specd_data.values()):
            log.warning(
                "Component %s is under-constrained, does not have any Physical types (e.g., value = 10kohm +/- 10%%).",
                component_addr)
    else:
        log.warning("No specification data provided for %s.", component_addr)

    # if there is no type, use the get_specd_type function to get it
    if "type" not in specd_data_dict:
        specd_data_dict["type"] = _get_specd_type(component_addr)

    # if there is a footprint, strip the leading letter and add the rest as 'package', remove the footprint
    if "footprint" in specd_data_dict:
        # if there is a package, use the package and remove the footprint
        type = _get_specd_type(component_addr)
        if "package" in specd_data_dict:
            del specd_data_dict["footprint"]
        # check if it is a resistor or capacitor
        elif type == "resistor" or type == "capacitor":
            specd_data_dict["package"] = specd_data_dict["footprint"][1:]
            del specd_data_dict["footprint"]


    cached_component = get_component_from_cache(component_addr, specd_data_dict)
    if cached_component:
        log.debug("Using cache for %s", component_addr)
        return cached_component

    url = config.get_project_context().config.services.components
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    try:
        response = requests.post(url, json=specd_data_dict, timeout=20, headers=headers)
        response.raise_for_status()
    except requests.HTTPError as ex:
        if ex.response.status_code == 404:
            friendly_dict = " && ".join(f"{k} == {v}" for k, v in specd_data_dict.items())
            raise NoMatchingComponent(
                f"No valid component found for spec {friendly_dict}: please check the part specs above, if they look right, we probably dont have it yet, we are working on it!",
                addr=component_addr
            ) from ex

        raise errors.AtoInfraError(
            f"""
            Failed to fetch component data from database.
            Error: {str(ex)}
            Response status code: {ex.response.status_code}
            Response text: {ex.response.text}
            """,
            addr=component_addr
        ) from ex
    except requests.RequestException as ex:
        raise errors.AtoInfraError(
            f"Error connecting to database: {str(ex)}",
            addr=component_addr
        ) from ex

    response_data = response.json() or {}

    best_component = response_data.get("bestComponent")
    # FIXME: Not returning something isn't a great mechanism to express
    # that we didn't find a component. It's not easy to distinguish between
    # a component not existing and other failure modes.
    if not best_component:
        raise NoMatchingComponent("No valid component found", addr=component_addr)

    lcsc = best_component["lcsc_id"]
    log.info("Fetched component %s for %s", lcsc, component_addr)

    # Now that we have a working component, update the cache with it for later
    update_cache(component_addr, best_component, specd_data_dict)

    return best_component


class MissingData(errors.AtoError):
    """
    Raised when a component is missing data in the Basic_Parts.csv file.
    """


# We cache the MPNs to ensure we select the same component if it's hit multiple times
# in a build
@cache
def get_mpn(addr: AddrStr) -> str:
    """
    Return the MPN for a component
    """
    specd_mpn = _get_specd_mpn(addr)
    if _is_generic(addr):
        return _get_generic_from_db(addr)["lcsc_id"]

    return specd_mpn


def get_specd_value(addr: AddrStr) -> str:
    """
    Return the MPN for a component given its address
    """
    comp_data = instance_methods.get_data_dict(addr)
    if not _is_generic(addr):
        # it's cool if there's no value for non-generics
        return str(comp_data.get("value", ""))

    try:
        return str(comp_data["value"])
    except KeyError as ex:
        raise MissingData(
            "$addr has no value spec'd",
            title="No value",
            addr=addr
        ) from ex


# Values come from the finally selected
# component, so we need to arbitrate via that data
@cache
def get_user_facing_value(addr: AddrStr) -> str:
    """
    Return a "value" of a component we can slap in things like
    the BoM and netlist. Doesn't need to be perfect, just
    something to look at.
    """
    if _is_generic(addr):
        db_data = _get_generic_from_db(addr)
        if db_data:
            return db_data.get("description", "")
        else:
            return "?"

    comp_data = instance_methods.get_data_dict(addr)
    # The default is okay here, because we're only generics
    # must have a value
    return str(comp_data.get("value", ""))


# FIXME: this function's requirements might cause a circular dependency
def download_footprint(addr: AddrStr, footprint_dir: Path):
    """
    Take the footprint from the database and make a .kicad_mod file for it
    TODO: clean this mess up
    """
    if not _is_generic(addr):
        return
    db_data = _get_generic_from_db(addr)

    # convert the footprint to a .kicad_mod file
    try:
        footprint = db_data.get("footprint_data", {})["kicad"]
    except KeyError as ex:
        raise MissingData(
            "db component for $addr has no footprint", title="No Footprint", addr=addr
        ) from ex

    if footprint == "standard_library":
        log.debug("Footprint is standard library, skipping")
        return

    try:
        file_name = db_data.get("footprint", {}).get("kicad")
        file_path = Path(footprint_dir) / str(file_name)

        # Create the directory if it doesn't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)

        with open(file_path, "w") as footprint_file:
            footprint_file.write(footprint)
    except Exception as ex:
        raise errors.AtoInfraError("Failed to write footprint file", addr=addr) from ex


# Footprints come from the users' code, so we reference that directly
@cache
def get_footprint(addr: AddrStr) -> str:
    """
    Return the footprint for a component
    """
    if _is_generic(addr):
        db_data = _get_generic_from_db(addr)

        try:
            footprint = db_data.get("footprint", {})["kicad"]
            # strip .kicad_mod from the end of the footprint if it's there
            if footprint.endswith(".kicad_mod"):
                footprint = footprint.removesuffix(".kicad_mod")
        except KeyError as ex:
            raise errors.AtoInfraError(
                "db component for $addr has no footprint",
                addr=addr
            ) from ex
        return footprint

    comp_data = instance_methods.get_data_dict(addr)
    try:
        return comp_data["footprint"]
    except KeyError as ex:
        raise MissingData(
            "$addr has no footprint", title="No Footprint", addr=addr
        ) from ex


@cache
def get_package(addr: AddrStr) -> str:
    """
    Return the package for a component
    """
    if _is_generic(addr):
        db_data = _get_generic_from_db(addr)
        return db_data.get("package", "")

    comp_data = instance_methods.get_data_dict(addr)
    try:
        return comp_data["footprint"]
    except KeyError as ex:
        raise MissingData("$addr has no package", title="No Package", addr=addr) from ex


class DesignatorManager:
    """TODO:"""

    def __init__(self) -> None:
        self._designators: dict[AddrStr, str] = {}

    def _make_designators(self, root: str) -> dict[str, str]:
        designators: dict[str, str] = {}
        unnamed_components = []
        used_designators = set()

        # FIXME: add lock-file data
        # first pass: grab all the designators from the lock data
        for component in filter(
            instance_methods.match_components, instance_methods.all_descendants(root)
        ):
            # designator = instance_methods.get_lock_data_dict(component).get(
            #     "designator"
            # )
            designator = None
            if designator:
                used_designators.add(designator)
                designators[component] = designator
            else:
                unnamed_components.append(component)

        # second pass: assign designators to the unnamed components
        for component in unnamed_components:
            prefix = instance_methods.get_data_dict(component).get(
                "designator_prefix", "U"
            )

            i = 1
            while f"{prefix}{i}" in used_designators:
                i += 1

            designators[component] = f"{prefix}{i}"
            used_designators.add(designators[component])
            # instance_methods.get_lock_data_dict(component)["designator"] = designators[
            #     component
            # ]

        return designators

    def get_designator(self, addr: str) -> str:
        """Return a mapping of instance address to designator."""
        if addr not in self._designators:
            self._designators = self._make_designators(address.get_entry(addr))
        return self._designators[addr]


designator_manager = DesignatorManager()


def get_designator(addr: str) -> str:
    """
    Return the designator for a component
    """
    return designator_manager.get_designator(addr)



import logging
from pathlib import Path

import pcbnew

from .common import get_layout_map, sync_footprints, flip_dict, sync_track

log = logging.getLogger(__name__)


class PullGroup(pcbnew.ActionPlugin):
    def defaults(self):
        self.name = "Pull Group"
        self.category = "Pull Group Layout Atopile"
        self.description = "Layout components on PCB in same spatial relationships as components on schematic"
        self.show_toolbar_button = True
        self.icon_file_name = str(Path(__file__).parent / "download.png")
        self.dark_icon_file_name = self.icon_file_name

    def Run(self):
        target_board: pcbnew.BOARD = pcbnew.GetBoard()
        board_path = target_board.GetFileName()
        known_layouts = get_layout_map(board_path)

        # Pull Selected Groups
        for g in target_board.Groups():
            assert isinstance(g, pcbnew.PCB_GROUP)

            # Skip over unselected groups
            if not g.IsSelected():
                continue

            g_name = g.GetName()
            if g_name not in known_layouts:
                continue

            layout_path = Path(known_layouts[g_name]["layout_path"])
            # Check what layouts exist
            if not layout_path.exists():
                raise FileNotFoundError(
                    f"Cannot load group. Layout file {layout_path} does not exist"
                )

            # Remove everything but the footprints from the target group
            for item in g.GetItems():
                if not isinstance(item, pcbnew.FOOTPRINT):
                    target_board.Remove(item)

            # Load the layout and sync
            source_board: pcbnew.BOARD = pcbnew.LoadBoard(str(layout_path))
            sync_footprints(
                source_board, target_board, flip_dict(known_layouts[g_name]["uuid_map"])
            )

            for track in source_board.GetTracks():
                item = sync_track(track, target_board)
                g.AddItem(item)

        pcbnew.Refresh()


PullGroup().register()



"""
This script largely controls the KiCAD CLI to generate
gerbers/drill files/etc... required to make circuit boards
"""

import logging
import re
import subprocess
import sys
import zipfile
from os import PathLike
from pathlib import Path
from time import time
from typing import Optional

import git

import atopile.errors
from atopile import config

log = logging.getLogger(__name__)


def find_kicad_cli() -> PathLike:
    """Figure out what to call for the KiCAD CLI."""
    if sys.platform.startswith("darwin"):
        kicad_cli_candidates = list(Path("/Applications/KiCad/").glob("**/kicad-cli"))
        # FIXME: handle multiple candidates
        return kicad_cli_candidates[0]
    elif sys.platform.startswith("linux"):
        return "kicad-cli"  # assume it's on the PATH


def run(*args, timeout_time: Optional[float] = 10, **kwargs) -> None:
    """Run a subprocess"""
    process = subprocess.Popen(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
        **kwargs
    )

    def _do_logging():
        outs, errs = process.communicate(timeout=0.1)
        for line in outs.splitlines():
            log.info(line)
        for line in errs.splitlines():
            log.error(line)

    start_time = time()
    while timeout_time is not None and time() - start_time < timeout_time:
        if process.poll() is not None:
            break
        try:
            _do_logging()
        except subprocess.TimeoutExpired:
            continue
    else:
        process.kill()

    exit_code = process.wait()
    _do_logging()

    if exit_code != 0:
        raise atopile.errors.AtoError(
            f"Command {args} failed with exit code {exit_code}"
        )


def generate_manufacturing_data(build_ctx: config.BuildContext) -> None:
    """Generate manufacturing data for the project."""
    # If there's no layout, we can't generate manufacturing data
    if not build_ctx.layout_path:
        atopile.errors.AtoError(
            "Layout must be available to generate manufacturing data"
        ).log(log, logging.WARNING)
        return

    # Ensure the build directory exists
    build_ctx.build_path.mkdir(parents=True, exist_ok=True)

    # Replace constants in the board file
    project_path = config.get_project_context().project_path
    try:
        repo = git.Repo(project_path, search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        atopile.errors.AtoError(
            "Project is not a git repository"
        ).log(log, logging.WARNING)
        short_githash = "nogit"
    else:
        short_githash_length = 7
        if repo.is_dirty():
            short_githash = "dirty"
        else:
            short_githash = repo.head.commit.hexsha[:short_githash_length]

    modded_kicad_pcb = build_ctx.output_base.with_suffix(
        ".kicad_pcb"
    )
    githash_kw = re.compile(re.escape("{{GITHASH}}"))
    with build_ctx.layout_path.open("r") as f_src, modded_kicad_pcb.open("w") as f_dst:
        for line in f_src:
            f_dst.write(githash_kw.sub(short_githash, line))

    # Setup for Gerbers
    gerber_dir = build_ctx.output_base.with_name(f"{build_ctx.output_base.name}-gerbers-{short_githash}")
    gerber_dir.mkdir(exist_ok=True, parents=True)
    gerber_dir_str = str(gerber_dir)
    if not gerber_dir_str.endswith("/"):
        gerber_dir_str += "/"

    kicad_cli = find_kicad_cli()

    # Generate Gerbers
    run([
        kicad_cli,
        "pcb",
        "export",
        "gerbers",
        "-o",
        gerber_dir_str,
        str(modded_kicad_pcb),
    ])
    run([
        kicad_cli,
        "pcb",
        "export",
        "drill",
        "-o",
        gerber_dir_str,
        str(modded_kicad_pcb),
    ])

    # Zip Gerbers
    zip_path = gerber_dir.with_suffix(".zip")
    with zipfile.ZipFile(zip_path, "w") as zip_file:
        for file in gerber_dir.glob("*"):
            zip_file.write(file)

    # Position files need some massaging for JLCPCB
    # We just need to replace the first row
    pos_path = build_ctx.output_base.with_suffix(".pos.csv")
    run([
        kicad_cli,
        "pcb",
        "export",
        "pos",
        "--format",
        "csv",
        "--units",
        "mm",
        "--use-drill-file-origin",
        "-o",
        str(pos_path),
        str(modded_kicad_pcb),
    ])
    pos_contents = pos_path.read_text().splitlines()
    pos_contents[0] = "Designator,Value,Package,Mid X,Mid Y,Rotation,Layer"
    pos_path.write_text("\n".join(pos_contents))



"""
Utils related to handling the parse tree
"""
from antlr4 import InputStream, Token, ParserRuleContext


def get_src_info_from_token(token: Token) -> tuple[str, int, int]:
    """Get the source path, line, and column from a context"""
    input_stream: InputStream = token.getInputStream()
    return input_stream.name, token.line, token.column


def get_src_info_from_ctx(ctx: ParserRuleContext) -> tuple[str, int, int]:
    """Get the source path, line, and column from a context"""
    token: Token = ctx.start
    return get_src_info_from_token(token)



import itertools
import logging
import os
import re
import shutil
import stat
import sys
import tempfile
import textwrap
import webbrowser
from pathlib import Path
from typing import Iterator, Optional

import caseconverter
import click
import git
import jinja2
import rich
import ruamel.yaml

from atopile import config, errors
from atopile.cli.install import install_core

# Set up logging
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


PROJECT_TEMPLATE = "https://github.com/atopile/project-template"


def check_name(name: str) -> bool:
    """
    Check if a name is valid.
    """
    if re.match(r"^[a-zA-Z][a-zA-Z0-9_-]*$", name):
        return True
    else:
        return False


def help(text: str) -> None:  # pylint: disable=redefined-builtin
    """Print help text."""
    rich.print("\n" + textwrap.dedent(text).strip() + "\n")


def _stuck_user_helper() -> Iterator[bool]:
    """Figure out if a user is stuck and help them exit."""
    threshold = 5
    for i in itertools.count():
        if i >= threshold:
            if rich.prompt.Confirm.ask("Are you trying to exit?"):
                rich.print("No worries! Try Ctrl+C next time!")
                exit(0)
            threshold += 5
        yield True


stuck_user_helper_generator = _stuck_user_helper()


@click.group()
def dev():
    pass



def _robustly_rm_dir(path: Path) -> None:
    """Remove a directory and all its contents."""
    def remove_readonly(func, path, excinfo):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onerror=remove_readonly)


def _in_git_repo(path: Path) -> bool:
    """Check if the current directory is in a git repo."""
    try:
        git.Repo(path)
    except git.InvalidGitRepositoryError:
        return False
    return True


@dev.command()
@click.argument("name", required=False)
@click.option("-r", "--repo", default=None)
def create(
    name: Optional[str],
    repo: Optional[str],
):  # pylint: disable=redefined-builtin
    """
    Create a new ato project or build configuration.
    """
    type = rich.prompt.Prompt.ask("What do you want to create", choices=["project", "build"])

    if type == "build":
        create_build()
        return

    # Get a project name
    kebab_name = None
    for _ in stuck_user_helper_generator:
        if not name:
            name = rich.prompt.Prompt.ask(
                ":rocket: What's your project [cyan]name?[/]", default=kebab_name
            )

        if name is None:
            continue

        kebab_name = caseconverter.kebabcase(name)
        if name != kebab_name:
            help(
                f"""
                We recommend using kebab-case ([cyan]{kebab_name}[/]) for your project name.
                It makes it easier to use your project with other tools (like git) and it embeds nicely into URLs.
                """
            )

            if rich.prompt.Confirm.ask(
                f"Do you want to use [cyan]{kebab_name}[/] instead?", default=True
            ):
                name = kebab_name

        if check_name(name):
            break
        else:
            help(
                "[red]Project names must start with a letter and"
                " contain only letters, numbers, dashes and underscores.[/]"
            )
            name = None

    if not repo and not rich.prompt.Confirm.ask(
        "Would you like to create a new repo for this project?"
    ):
        repo = PROJECT_TEMPLATE

    # Get a repo
    for _ in stuck_user_helper_generator:
        if not repo:
            make_repo_url = f"https://github.com/new?name={name}&template_owner=atopile&template_name=project-template"

            help(
                f"""
                We recommend you create a Github repo for your project.

                If you already have a repo, you can respond [yellow]n[/]
                to the next question and provide the URL to your repo.

                If you don't have one, you can respond yes to the next question
                or (Cmd/Ctrl +) click the link below to create one.

                Just select the template you want to use.

                {make_repo_url}
                """
            )
            if rich.prompt.Confirm.ask(
                ":rocket: Open browser to create Github repo?", default=True
            ):
                webbrowser.open(make_repo_url)

            repo = rich.prompt.Prompt.ask(":rocket: What's the [cyan]repo's URL?[/]")

        # Try download the repo from the user-provided URL
        if Path(name).exists():
            raise click.ClickException(
                f"Directory {name} already exists. Please put the repo elsewhere or choose a different name."
            )

        try:
            repo_obj = git.Repo.clone_from(repo, name, depth=1)
            break
        except git.GitCommandError as ex:
            help(
                f"""
                [red]Failed to clone repo from {repo}[/]

                {ex.stdout}
                {ex.stderr}
                """
            )
            repo = None

    # Configure the project
    do_configure(name, repo_obj.working_tree_dir, debug=False)

    # Commit the configured project
    # force the add, because we're potentially
    # modifying things in gitignored locations
    if repo_obj.is_dirty():
        repo_obj.git.add(A=True, f=True)
        repo_obj.git.commit(m="Configure project")
    else:
        rich.print(
            "[yellow]No changes to commit! Seems like the"
            " template you used mightn't be configurable?[/]"
        )

    # If this repo's remote it PROJECT_TEMPLATE, cleanup the git history
    if repo_obj.remotes.origin.url == PROJECT_TEMPLATE:
        try:
            _robustly_rm_dir(repo_obj.git_dir)
        except (PermissionError, OSError) as ex:
            errors.AtoError(
                f"Failed to remove .git directory: {repr(ex)}"
            ).log(log, logging.WARNING)
        if not _in_git_repo(Path(repo_obj.working_dir).parent):
            # If we've created this project OUTSIDE an existing git repo
            # then re-init the repo so it has a clean history
            clean_repo = git.Repo.init(repo_obj.working_tree_dir)
            clean_repo.git.add(A=True)
            clean_repo.git.commit(m="Initial commit")

    # Install dependencies listed in the ato.yaml, typically just generics
    install_core(
        to_install="", jlcpcb=False, upgrade=True, path=repo_obj.working_tree_dir
    )

    # Wew! New repo created!
    rich.print(f':sparkles: [green]Created new project "{name}"![/] :sparkles:')


def create_build():
    """
    Create a new build configuration.
    - adds entry to ato.yaml
    - creates a new directory in layout
    """
    build_name = caseconverter.kebabcase(rich.prompt.Prompt.ask("Enter the build name"))
    try:
        project_config = config.get_project_config_from_path(Path("."))
        project_context = config.ProjectContext.from_config(project_config)
        top_level_path = project_context.project_path
        layout_path = project_context.layout_path
        src_path = project_context.src_path
    except FileNotFoundError:
        raise errors.AtoError("Could not find the project directory, are you within an ato project?")

    # Get user input for the entry file and module name
    rich.print("We will create a new ato file and add the entry to the ato.yaml")
    entry = rich.prompt.Prompt.ask("What would you like to call the entry file? (e.g., psuDebug)")

    target_layout_path = layout_path / build_name
    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            git.Repo.clone_from(PROJECT_TEMPLATE, tmpdirname)
        except git.GitCommandError as ex:
            raise errors.AtoError(f"Failed to clone layout template from {PROJECT_TEMPLATE}: {repr(ex)}")
        source_layout_path = Path(tmpdirname) / "elec" / "layout" / "default"
        if not source_layout_path.exists():
            raise errors.AtoError(f"The specified layout path {source_layout_path} does not exist.")
        else:
            target_layout_path.mkdir(parents=True, exist_ok=True)
            shutil.copytree(source_layout_path, target_layout_path, dirs_exist_ok=True)
            # Configure the files in the directory using the do_configure function
            do_configure(build_name, str(target_layout_path), debug=False)

        # Add the build to the ato.yaml file
        ato_yaml_path = top_level_path / config.CONFIG_FILENAME
        # Check if ato.yaml exists
        if not ato_yaml_path.exists():
            print(f"ato.yaml not found in {top_level_path}. Please ensure the file exists before proceeding.")
        else:
            # Load the existing YAML configuration
            yaml = ruamel.yaml.YAML()
            with ato_yaml_path.open("r") as file:
                ato_config = yaml.load(file)

            entry_file = Path(caseconverter.kebabcase(entry)).with_suffix(".ato")
            entry_module = caseconverter.pascalcase(entry)

            # Update the ato_config with the new build information
            if "builds" not in ato_config:
                ato_config["builds"] = {}
            ato_config["builds"][build_name] = {
                "entry": f"elec/src/{entry_file}:{entry_module}"
            }

            # Write the updated configuration back to ato.yaml
            with ato_yaml_path.open("w") as file:
                yaml.dump(ato_config, file)


        # create a new ato file with the entry file and module
        ato_file = src_path / entry_file

        ato_file.write_text(f"module {entry_module}:\n \tsignal gnd\n")

        rich.print(f":sparkles: Successfully created a new build configuration for {build_name}! :sparkles:")


@dev.command()
@click.argument("name")
@click.argument("repo_path")
def configure(name: str, repo_path: str):
    """Command useful in developing templates."""
    do_configure(name, repo_path, debug=True)


def do_configure(name: str, _repo_path: str, debug: bool):
    """Configure the project."""
    repo_path = Path(_repo_path)
    try:
        author = git.Repo(repo_path).git.config("user.name")
    except (git.GitCommandError, git.InvalidGitRepositoryError):
        author = "Original Author"

    template_globals = {
        "name": name,
        "caseconverter": caseconverter,
        "repo_root": repo_path,
        "python_path": sys.executable,
        "author": author,
    }

    # Load templates
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(repo_path)))

    for template_path in repo_path.glob("**/*.j2"):
        # Figure out the target path and variables and what not
        target_path = template_path.parent / template_path.name.replace(
            ".j2", ""
        ).replace("__name__", caseconverter.kebabcase(name))

        template_globals["rel_path"] = target_path

        template = env.get_template(
            str(template_path.relative_to(repo_path).as_posix()),
            globals=template_globals,
        )

        # Make the noise!
        with target_path.open("w") as f:
            for chunk in template.generate():
                f.write(chunk)

        # Remove the template
        if not debug:
            template_path.unlink()


if __name__ == "__main__":
    dev()  # pylint: disable=no-value-for-parameter



from typing import Any, Callable, Iterable, Optional

from atopile import address, errors
from atopile.address import AddrStr
from atopile.front_end import ClassLayer, Link, lofty, Instance


def get_children(addr: str) -> Iterable[AddrStr]:
    """Return the children of a given address."""
    instance = lofty.get_instance(addr)
    for child in instance.children.values():
        yield child.addr


def get_data_dict(addr: str) -> dict[str, Any]:
    """Return the data at the given address"""
    instance = lofty.get_instance(addr)
    return {k: v[0].value for k, v in instance.assignments.items()}


def all_descendants(addr: str) -> Iterable[str]:
    """
    Return a list of addresses in depth-first order
    """
    for child in get_children(addr):
        yield from all_descendants(child)
    yield addr


def _make_dumb_matcher(pass_list: Iterable[str]) -> Callable[[str], bool]:
    """
    Return a filter that checks if the addr is in the pass_list
    """

    def _filter(addr: AddrStr) -> bool:
        instance = lofty.get_instance(addr)
        for super_ in reversed(instance.supers):
            if super_.address in pass_list:
                return True
        return False

    return _filter


match_components = _make_dumb_matcher(["<Built-in>:Component"])
match_modules = _make_dumb_matcher(["<Built-in>:Module"])
match_signals = _make_dumb_matcher(["<Built-in>:Signal"])
match_pins = _make_dumb_matcher("<Built-in>:Pin")
match_pins_and_signals = _make_dumb_matcher(["<Built-in>:Pin", "<Built-in>:Signal"])
match_interfaces = _make_dumb_matcher(["<Built-in>:Interface"])
match_sentinels = _make_dumb_matcher(
    [
        "<Built-in>:Component",
        "<Built-in>:Module" "<Built-in>:Signal",
        "<Built-in>:Pin",
        "<Built-in>:Interface",
    ]
)


def find_matching_super(
    addr: AddrStr, candidate_supers: list[AddrStr]
) -> Optional[AddrStr]:
    """
    Return the first super of addr, is in the list of
    candidate_supers, or None if none are.
    """
    supers = get_supers_list(addr)
    for duper in supers:
        if any(duper.address == pt for pt in candidate_supers):
            return duper.address
    return None


def get_supers_list(addr: AddrStr) -> list[ClassLayer]:
    """Return the supers of an object as a list of ObjectLayers."""
    return lofty.get_instance(addr).supers


def get_next_super(addr: AddrStr) -> ClassLayer:
    """Return the immediate super-class of the given address."""
    return get_supers_list(addr)[0]


def get_parent(addr: AddrStr) -> Optional[AddrStr]:
    """
    Return the parent of the given address
    """
    instance = lofty.get_instance(addr).parent
    if instance:
        return instance.addr
    return None


def iter_parents(addr: AddrStr) -> Iterable[AddrStr]:
    """Iterate over the parents of the given address"""
    while addr := get_parent(addr):
        yield addr


def get_links(addr: AddrStr) -> Iterable[Link]:
    """Return the links associated with an instance"""
    yield from lofty.get_instance(addr).links


def get_instance(addr: AddrStr) -> Instance:
    """Return the instance at the given address"""
    inst = lofty.get_instance(address.get_entry(addr))
    inst_names = address.get_instance_names(addr)
    for inst_name in inst_names:
        try:
            inst = inst.children[inst_name]
        except KeyError as ex:
            raise errors.AtoKeyError(
                f"Instance {address.add_instance(inst.addr, inst_name)} not found"
            ) from ex
    return inst



import operator
from collections.abc import Callable

from atopile.expressions import Expression, defer_operation_factory


def test_two_callables():
    a = Expression(symbols=set(), lambda_=lambda ctx: ctx["d"])
    b = Expression(symbols=set(), lambda_=lambda ctx: 21)

    c = defer_operation_factory(a, operator.add, b)

    assert isinstance(c, Callable)
    assert c({"d": 157}) == 178


def test_one_callable():
    a = Expression(symbols=set(), lambda_=lambda ctx: ctx["d"])
    b = 12

    c = defer_operation_factory(a, operator.add, b)

    assert isinstance(c, Callable)
    assert c({"d": 58}) == 70


def test_no_callables():
    a = 12
    b = 21

    c = defer_operation_factory(a, operator.add, b)

    assert not isinstance(c, Callable)
    assert c == 33



import logging

import click
from rich.logging import RichHandler

from atopile import telemetry
from atopile.cli.rich_console import console

from . import build, configure, create, inspect, install

FORMAT = "%(message)s"
logging.basicConfig(
    level="INFO",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[
        RichHandler(
            console=console,
            tracebacks_suppress=[click],
        )
    ],
)


# cli root
@click.version_option()
@click.group()
@click.option("--non-interactive", is_flag=True, envvar="ATO_NON_INTERACTIVE")
@click.option("--debug", is_flag=True)
@click.option("-v", "--verbose", count=True)
@click.pass_context  # This decorator makes the context available to the command.
def cli(ctx, non_interactive: bool, debug: bool, verbose: int):
    """Base CLI group."""

    # Initialize telemetry
    telemetry.setup_telemetry_data(ctx.invoked_subcommand)

    if debug:
        import debugpy  # pylint: disable=import-outside-toplevel

        debug_port = 5678
        debugpy.listen(("localhost", debug_port))
        logging.info("Starting debugpy on port %s", debug_port)
        debugpy.wait_for_client()

    # set the log level
    if verbose == 1:
        logging.root.setLevel(logging.DEBUG)
    elif verbose > 1:
        logging.root.setLevel(logging.NOTSET)

    if not non_interactive:
        configure.do_configure_if_needed()


cli.add_command(build.build)
cli.add_command(create.create)
cli.add_command(install.install)
cli.add_command(configure.configure)
cli.add_command(inspect.inspect)


if __name__ == "__main__":
    cli()



"""
This datamodel represents the code in a clean, simple and traversable way,
but doesn't resolve names of things.
In building this datamodel, we check for name collisions, but we don't resolve them yet.
"""

import enum
import operator
from collections import defaultdict, deque
from contextlib import ExitStack, contextmanager
from itertools import chain
from numbers import Number
from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Optional

import pint
from antlr4 import ParserRuleContext
from attrs import define, field, resolve_types

from atopile import address, config, errors, expressions
from atopile.address import AddrStr
from atopile.datatypes import KeyOptItem, KeyOptMap, Ref, StackList
from atopile.generic_methods import recurse
from atopile.parse import parser
from atopile.parse_utils import get_src_info_from_ctx
from atopile.parser.AtopileParser import AtopileParser as ap
from atopile.parser.AtopileParserVisitor import AtopileParserVisitor


@define
class Base:
    """Represent a base class for all things."""

    src_ctx: Optional[ParserRuleContext] = field(kw_only=True, default=None)


@define
class Import(Base):
    """Represent an import statement."""

    obj_addr: AddrStr

    def __repr__(self) -> str:
        return f"<Import {self.obj_addr}>"


@define
class Replacement(Base):
    """Represent a replacement statement."""

    new_super_ref: Ref


@define(repr=False)
class ClassDef(Base):
    """
    Represent the definition or skeleton of an object
    so we know where we can go to find the object later
    without actually building the whole file.

    This is mainly because we don't want to hit errors that
    aren't relevant to the current build - instead leaving them
    to be hit in the case we're actually building that object.
    """

    super_ref: Optional[Ref]
    imports: Mapping[Ref, Import]

    local_defs: Mapping[Ref, "ClassDef"]
    replacements: Mapping[Ref, Replacement]

    # attached immediately to the object post construction
    closure: Optional[tuple["ClassDef"]] = None  # in order of lookup
    address: Optional[AddrStr] = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.address}>"


class RangedValue(expressions.RangedValue):
    def __init__(
        self,
        val_a: Number | pint.Quantity,
        val_b: Number | pint.Quantity,
        unit: Optional[pint.Unit] = None,
        src_ctx: Optional[ParserRuleContext] = None,
    ):
        self.src_ctx = src_ctx
        super().__init__(val_a, val_b, unit)


class Expression(expressions.Expression):
    def __init__(
        self,
        symbols: set[str],
        lambda_: Callable,
        src_ctx: Optional[ParserRuleContext] = None
    ):
        super().__init__(symbols=symbols, lambda_=lambda_)
        self.src_ctx = src_ctx

    def __repr__(self) -> str:
        if not self.src_ctx:
            return f"<{self.__class__.__name__}>"

        return f"<{self.__class__.__name__} {str(self)}>"

    def __str__(self) -> str:
        return self.src_ctx.getText()


class Assertion:
    def __init__(
        self,
        lhs: expressions.Expression,
        operator: str,
        rhs: expressions.Expression,
        src_ctx: Optional[ParserRuleContext] = None,
    ):
        self.lhs = lhs
        self.operator = operator
        self.rhs = rhs
        self.src_ctx = src_ctx

    def __str__(self) -> str:
        return f"{self.lhs} {self.operator} {self.rhs}"


_dimensionality_to_unit_map = {
    "None": pint.Unit("dimensionless"),
    "length": pint.Unit("m"),
    "time": pint.Unit("s"),
    "voltage": pint.Unit("V"),
    "current": pint.Unit("A"),
    "resistance": pint.Unit("ohm"),
    "capacitance": pint.Unit("F"),
    "inductance": pint.Unit("H"),
    "frequency": pint.Unit("Hz"),
    "power": pint.Unit("W"),
    "energy": pint.Unit("J"),
    "charge": pint.Unit("C"),
}


@define
class Assignment(Base):
    """Represent an assignment statement."""

    name: str
    value: Optional[Any]
    given_type: Optional[str | pint.Unit]

    @property
    def unit(self) -> pint.Unit:
        """Return a pint unit from an assignment, or raise an exception if there's no unit."""
        if self.given_type is None:
            # Handle an implicit type
            try:
                return self.value.unit
            except AttributeError as ex:
                raise errors.AtoTypeError(
                    f"Assignment '{self.name}' has no unit"
                ) from ex
        try:
            return _dimensionality_to_unit_map[self.given_type]
        except KeyError as ex:
            raise errors.AtoUnknownUnitError(
                f"Unknown dimensionality '{self.given_type}'"
            ) from ex

    def __repr__(self) -> str:
        return f"<Assignment {self.name} = {self.value}>"


@define(repr=False)
class ClassLayer(Base):
    """
    Represent a layer in the object hierarchy.
    This holds all the values assigned to the object.
    """

    # information about where this object is found in multiple forms
    # this is redundant with one another (eg. you can compute one from the other)
    # but it's useful to have all of them for different purposes
    obj_def: ClassDef

    # None indicates that this is a root object
    super: Optional["ClassLayer"]

    # the local objects and vars are things we navigate to a lot
    assignments: Mapping[str, Assignment]

    @property
    def address(self) -> AddrStr:
        return self.obj_def.address

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.obj_def.address}>"


resolve_types(ClassLayer)


## The below datastructures are created from the above data-model as a second stage


@define
class Link(Base):
    """Represent a connection between two connectable things."""

    # TODO: we may not need this using loop-soup
    # the reason this currently exists is to allow us to map joints between instances
    # these make sense only in the context of the pins and signals, which aren't
    # language fundamentals as much as net objects - eg. they're useful only from
    # a specific electrical perspective
    # origin_link: Link

    parent: "Instance"
    source: "Instance"
    target: "Instance"

    def __repr__(self) -> str:
        return f"<Link {repr(self.source)} -> {repr(self.target)}>"


@define
class Instance(Base):
    """
    Represents the specific instance, capturing, the story you told of
    how to get there in it's mapping stacks.
    """

    # origin information
    # much of this information is redundant, however it's all highly referenced
    # so it's useful to have it all at hand
    addr: AddrStr
    supers: list["ClassLayer"]
    assignments: Mapping[str, deque[Assignment]]
    parent: Optional["Instance"]

    # created as supers' ASTs are walked
    assertions: list[Assertion] = field(factory=list)
    children: dict[str, "Instance"] = field(factory=dict)
    links: list[Link] = field(factory=list)

    def __repr__(self) -> str:
        return f"<Instance {self.addr}>"

    @classmethod
    def from_super(
        cls,
        addr: AddrStr,
        super_: ClassLayer,
        parent: Optional["Instance"],
        src_ctx: Optional[ParserRuleContext] = None,
    ) -> "Instance":
        """Create an instance from a list of supers."""
        supers = list(recurse(lambda x: x.super, super_))

        assignments = defaultdict(deque)

        for super_ in supers:
            for k, v in super_.assignments.items():
                assignments[k].append(v)

        return cls(
            src_ctx=src_ctx,
            addr=addr,
            supers=supers,
            assignments=assignments,
            parent=parent,
        )


resolve_types(Instance)
resolve_types(Link)


class _Sentinel(enum.Enum):
    NOTHING = enum.auto()


NOTHING = _Sentinel.NOTHING


def _make_obj_layer(address: AddrStr, super: Optional[ClassLayer] = None) -> ClassLayer:
    """Create a new object layer from an address and a set of supers."""
    obj_def = ClassDef(
        address=address,
        super_ref=Ref.empty(),
        imports={},
        local_defs={},
        replacements={},
    )
    return ClassLayer(
        obj_def=obj_def,
        super=super,
        assignments={},
    )


MODULE: ClassLayer = _make_obj_layer(AddrStr("<Built-in>:Module"))
COMPONENT: ClassLayer = _make_obj_layer(AddrStr("<Built-in>:Component"), super=MODULE)
PIN: ClassLayer = _make_obj_layer(AddrStr("<Built-in>:Pin"))
SIGNAL: ClassLayer = _make_obj_layer(AddrStr("<Built-in>:Signal"))
INTERFACE: ClassLayer = _make_obj_layer(AddrStr("<Built-in>:Interface"))


BUILTINS_BY_REF = {
    Ref.from_one("MODULE"): MODULE,
    Ref.from_one("COMPONENT"): COMPONENT,
    Ref.from_one("INTERFACE"): INTERFACE,
    Ref.from_one("PIN"): PIN,
    Ref.from_one("SIGNAL"): SIGNAL,
}


BUILTINS_BY_ADDR = {
    MODULE.address: MODULE,
    COMPONENT.address: COMPONENT,
    PIN.address: PIN,
    SIGNAL.address: SIGNAL,
    INTERFACE.address: INTERFACE,
}


def _get_unit_from_ctx(ctx: ParserRuleContext) -> pint.Unit:
    """Return a pint unit from a context."""
    unit_str = ctx.getText()
    try:
        return pint.Unit(unit_str)
    except pint.UndefinedUnitError as ex:
        raise errors.AtoUnknownUnitError.from_ctx(
            ctx, f"Unknown unit '{unit_str}'"
        ) from ex


class HandlesPrimaries(AtopileParserVisitor):
    """
    This class is a mixin to be used with the translator classes.
    """
    def visit_ref_helper(
        self,
        ctx: (
            ap.NameContext
            | ap.AttrContext
            | ap.Name_or_attrContext
            | ap.Totally_an_integerContext
        ),
    ) -> Ref:
        """
        Visit any referencey thing and ensure it's returned as a reference
        """
        return Ref(ctx.getText().split("."))

    def visitName_or_attr(self, ctx: ap.Name_or_attrContext) -> Ref:
        if ctx.name():
            name = self.visitName(ctx.name())
            return Ref.from_one(name)
        elif ctx.attr():
            return self.visitAttr(ctx.attr())

        raise errors.AtoError("Expected a name or attribute")

    def visitName(self, ctx: ap.NameContext) -> str:
        """
        If this is an int, convert it to one (for pins), else return the name as a string.
        """
        return ctx.getText()

    def visitAttr(self, ctx: ap.AttrContext) -> Ref:
        return Ref(self.visitName(name) for name in ctx.name())

    def visitString(self, ctx: ap.StringContext) -> str:
        return ctx.getText().strip("\"'")

    def visitBoolean_(self, ctx: ap.Boolean_Context) -> bool:
        return ctx.getText().lower() == "true"

    def visitLiteral_physical(self, ctx: ap.Literal_physicalContext) -> RangedValue:
        """Yield a physical value from a physical context."""
        if ctx.implicit_quantity():
            return self.visitImplicit_quantity(ctx.implicit_quantity())
        if ctx.bilateral_quantity():
            return self.visitBilateral_quantity(ctx.bilateral_quantity())
        if ctx.bound_quantity():
            return self.visitBound_quantity(ctx.bound_quantity())

        raise ValueError  # this should be protected because it shouldn't be parseable

    def visitImplicit_quantity(self, ctx: ap.Implicit_quantityContext) -> RangedValue:
        """Yield a physical value from an implicit quantity context."""
        value = float(ctx.NUMBER().getText())

        if ctx.name():
            unit = _get_unit_from_ctx(ctx.name())
        else:
            unit = pint.Unit("")

        return RangedValue(
            src_ctx=ctx,
            val_a=value,
            val_b=value,
            unit=unit,
        )

    def visitBilateral_quantity(self, ctx: ap.Bilateral_quantityContext) -> RangedValue:
        """Yield a physical value from a bilateral quantity context."""
        nominal = float(ctx.bilateral_nominal().NUMBER().getText())

        if ctx.bilateral_nominal().name():
            unit = _get_unit_from_ctx(ctx.bilateral_nominal().name())
        else:
            unit = pint.Unit("")

        tol_ctx: ap.Bilateral_toleranceContext = ctx.bilateral_tolerance()
        tol_num = float(tol_ctx.NUMBER().getText())

        if tol_ctx.PERCENT():
            tol_divider = 100
        # FIXME: hardcoding this seems wrong, but the parser/lexer wasn't picking up on it
        elif tol_ctx.name() and tol_ctx.name().getText() == "ppm":
            tol_divider = 1e6
        else:
            tol_divider = None

        if tol_divider:
            if nominal == 0:
                raise errors.AtoError.from_ctx(
                    tol_ctx,
                    "Can't calculate tolerance percentage of a nominal value of zero",
                )

            # In this case, life's a little easier, and we can simply multiply the nominal
            return RangedValue(
                src_ctx=ctx,
                val_a=nominal * (1 - tol_num / tol_divider),
                val_b=nominal * (1 + tol_num / tol_divider),
                unit=unit,
            )

        if tol_ctx.name():
            # In this case there's a named unit on the tolerance itself
            # We need to make sure it's dimensionally compatible with the nominal
            tolerance_unit = _get_unit_from_ctx(tol_ctx.name())
            try:
                tolerance = (tol_num * tolerance_unit).to(unit).magnitude
            except pint.DimensionalityError as ex:
                raise errors.AtoTypeError.from_ctx(
                    tol_ctx.name(),
                    f"Tolerance unit '{tolerance_unit}' is not dimensionally"
                    f" compatible with nominal unit '{unit}'",
                ) from ex

            return RangedValue(
                src_ctx=ctx,
                val_a=nominal - tolerance,
                val_b=nominal + tolerance,
                unit=unit,
            )

        # If there's no unit or percent, then we have a simple tolerance in the same units
        # as the nominal
        return RangedValue(
            src_ctx=ctx,
            val_a=nominal - tol_num,
            val_b=nominal + tol_num,
            unit=unit,
        )

    def visitBound_quantity(self, ctx: ap.Bound_quantityContext) -> RangedValue:
        """Yield a physical value from a bound quantity context."""

        def _parse_end(
            ctx: ap.Quantity_endContext,
        ) -> tuple[float, Optional[pint.Unit]]:
            value = float(ctx.NUMBER().getText())
            if ctx.name():
                unit = _get_unit_from_ctx(ctx.name())
            else:
                unit = None
            return value, unit

        start_val, start_unit = _parse_end(ctx.quantity_end(0))
        end_val, end_unit = _parse_end(ctx.quantity_end(1))

        if start_unit is None and end_unit is None:
            unit = pint.Unit("")
        elif start_unit and end_unit:
            unit = start_unit
            try:
                end_val = (end_val * end_unit).to(start_unit).magnitude
            except pint.DimensionalityError as ex:
                raise errors.AtoTypeError.from_ctx(
                    ctx,
                    f"Tolerance unit '{end_unit}' is not dimensionally"
                    f" compatible with nominal unit '{start_unit}'",
                ) from ex
        elif start_unit:
            unit = start_unit
        else:
            unit = end_unit

        return RangedValue(
            src_ctx=ctx,
            val_a=start_val,
            val_b=end_val,
            unit=unit,
        )


class HandleStmtsFunctional(AtopileParserVisitor):
    """
    The base translator is responsible for methods common to
    navigating from the top of the AST including how to process
    errors, and commonising return types.
    """

    def defaultResult(self):
        """
        Override the default "None" return type
        (for things that return nothing) with the Sentinel NOTHING
        """
        return NOTHING

    def visit_iterable_helper(self, children: Iterable) -> KeyOptMap:
        """
        Visit multiple children and return a tuple of their results,
        discard any results that are NOTHING and flattening the children's results.
        It is assumed the children are returning their own OptionallyNamedItems.
        """

        def __visit():
            for err_cltr, child in errors.iter_through_errors(children):
                with err_cltr():
                    child_result = self.visit(child)
                    if child_result is not NOTHING:
                        yield child_result

        child_results = chain.from_iterable(__visit())
        child_results = list(item for item in child_results if item is not NOTHING)
        child_results = KeyOptMap(KeyOptItem(cr) for cr in child_results)

        return KeyOptMap(child_results)

    def visitSimple_stmt(
        self, ctx: ap.Simple_stmtContext
    ) -> Iterable[_Sentinel | KeyOptItem]:
        """
        This is practically here as a development shim to assert the result is as intended
        """
        result = self.visitChildren(ctx)
        for item in result:
            if item is not NOTHING:
                assert isinstance(item, KeyOptItem)
        return result

    def visitStmt(self, ctx: ap.StmtContext) -> KeyOptMap:
        """
        Ensure consistency of return type.
        We choose to raise any below exceptions here, because stmts can be nested,
        and raising exceptions serves as our collection mechanism.
        """
        if ctx.simple_stmts():
            stmt_returns = self.visitSimple_stmts(ctx.simple_stmts())
            return stmt_returns
        elif ctx.compound_stmt():
            item = self.visit(ctx.compound_stmt())
            if item is NOTHING:
                return KeyOptMap.empty()
            assert isinstance(item, KeyOptItem)
            return KeyOptMap.from_item(item)

        raise TypeError("Unexpected statement type")

    def visitSimple_stmts(self, ctx: ap.Simple_stmtsContext) -> KeyOptMap:
        return self.visit_iterable_helper(ctx.simple_stmt())

    def visitBlock(self, ctx) -> KeyOptMap:
        if ctx.stmt():
            return self.visit_iterable_helper(ctx.stmt())
        if ctx.simple_stmts():
            return self.visitSimple_stmts(ctx.simple_stmts())
        raise ValueError  # this should be protected because it shouldn't be parseable


# TODO: actually capture src_ctx on expressions
class Roley(HandlesPrimaries):
    """
    Roley is a green road roller who loves to make up songs and often ends his
    sentences with "Rock and Roll!" He is enthusiastic and works alongside Bob
    and the rest of the team on various construction projects. Roley's main job
    is to smooth out roads and pavements, making sure they are flat and safe
    for everyone. He takes great pride in creating perfect surfaces and is an
    essential member of Bob's team, helping to complete construction tasks with
    his unique abilities.

    Roley's also builds expressions.
    """
    def __init__(self, addr: AddrStr) -> None:
        self.addr = addr
        super().__init__()

    def visitArithmetic_expression(
        self, ctx: ap.Arithmetic_expressionContext
    ) -> expressions.NumericishTypes:
        if ctx.ADD():
            return expressions.defer_operation_factory(
                self.visit(ctx.arithmetic_expression()),
                operator.add,
                self.visit(ctx.term()),
            )

        if ctx.MINUS():
            return expressions.defer_operation_factory(
                self.visit(ctx.arithmetic_expression()),
                operator.sub,
                self.visit(ctx.term()),
            )

        return self.visit(ctx.term())

    def visitTerm(self, ctx: ap.TermContext) -> expressions.NumericishTypes:
        if ctx.STAR():  # multiply
            return expressions.defer_operation_factory(
                self.visit(ctx.term()),
                operator.mul,
                self.visit(ctx.factor()),
            )

        if ctx.DIV():
            return expressions.defer_operation_factory(
                self.visit(ctx.term()),
                operator.truediv,
                self.visit(ctx.factor()),
            )

        return self.visit(ctx.factor())

    def visitFactor(self, ctx: ap.FactorContext) -> expressions.NumericishTypes:
        # Ignore the unary plus operator

        if ctx.MINUS():
            return operator.neg(self.visit(ctx.power()))

        return self.visit(ctx.power())

    def visitPower(self, ctx: ap.PowerContext) -> expressions.NumericishTypes:
        if ctx.factor():
            return expressions.defer_operation_factory(
                self.visit(ctx.atom()),
                operator.pow,
                self.visit(ctx.factor()),
            )

        return self.visit(ctx.atom())

    def visitAtom(self, ctx: ap.AtomContext) -> expressions.NumericishTypes:
        if ctx.arithmetic_group():
            return self.visit(ctx.arithmetic_group().arithmetic_expression())

        if ctx.literal_physical():
            return self.visit(ctx.literal_physical())

        if ctx.name_or_attr():
            return expressions.Symbol(
                address.add_instances(
                    self.addr,
                    self.visit_ref_helper(ctx.name_or_attr()),
                )
            )

        raise ValueError


class Scoop(HandleStmtsFunctional, HandlesPrimaries):
    """Scoop's job is to map out all the object definitions in the code."""

    def __init__(
        self,
        ast_getter: Callable[[str | Path], ParserRuleContext],
    ) -> None:
        self.ast_getter = ast_getter
        self._output_cache: dict[AddrStr, ClassDef] = {}
        super().__init__()

    def get_search_paths(self) -> Iterable[Path]:
        """Return the search paths."""
        project_context = config.get_project_context()
        return [project_context.src_path, project_context.module_path]

    def get_obj_def(self, addr: AddrStr) -> ClassDef:
        """Returns the ObjectDef for a given address."""
        if addr not in self._output_cache:
            file = address.get_file(addr)
            file_ast = self.ast_getter(file)
            obj = self.visitFile_input(file_ast)
            assert isinstance(obj, ClassDef)
            # this operation puts it and it's children in the cache
            self._register_obj_tree(obj, AddrStr(file), ())
        try:
            return self._output_cache[addr]
        except KeyError as ex:
            raise BlockNotFoundError(
                f"No block named $addr in {address.get_file(addr)}", addr=addr
            ) from ex

    def _register_obj_tree(
        self, obj: ClassDef, addr: AddrStr, closure: tuple[ClassDef]
    ) -> None:
        """Register address info to the object, and add it to the cache."""
        obj.address = addr
        obj.closure = closure
        child_closure = (obj,) + closure
        self._output_cache[addr] = obj
        for ref, child in obj.local_defs.items():
            assert len(ref) == 1
            assert isinstance(ref[0], str)
            child_addr = address.add_entry(addr, ref[0])
            self._register_obj_tree(child, child_addr, child_closure)

    def visitFile_input(self, ctx: ap.File_inputContext) -> ClassDef:
        """Visit a file input and return it's object."""
        locals_ = self.visit_iterable_helper(ctx.stmt())

        # FIXME: clean this up, and do much better name collision detection on it
        local_defs = {}
        imports = {}
        for ref, local in locals_:
            if isinstance(local, ClassDef):
                local_defs[ref] = local
            elif isinstance(local, Import):
                assert ref is not None
                imports[ref] = local
            else:
                raise errors.AtoError(f"Unexpected local type: {type(local)}")

        file_obj = ClassDef(
            src_ctx=ctx,
            super_ref=Ref.from_one("MODULE"),
            imports=imports,
            local_defs=local_defs,
            replacements={},
        )

        return file_obj

    def visitBlockdef(self, ctx: ap.BlockdefContext) -> KeyOptItem[ClassDef]:
        """Visit a blockdef and return it's object."""
        if ctx.FROM():
            if not ctx.name_or_attr():
                raise errors.AtoSyntaxError("Expected a name or attribute after 'from'")
            block_super_ref = self.visit_ref_helper(ctx.name_or_attr())
        else:
            block_super_ref = self.visitBlocktype(ctx.blocktype())

        locals_ = self.visitBlock(ctx.block())

        # FIXME: this needs far better name collision detection
        local_defs = {}
        imports = {}
        replacements = {}
        for ref, local in locals_:
            if isinstance(local, ClassDef):
                local_defs[ref] = local
            elif isinstance(local, Import):
                imports[ref] = local
            elif isinstance(local, Replacement):
                replacements[ref] = local
            else:
                raise errors.AtoError(f"Unexpected local type: {type(local)}")

        block_obj = ClassDef(
            src_ctx=ctx,
            super_ref=block_super_ref,
            imports=imports,
            local_defs=local_defs,
            replacements=replacements,
        )

        block_name = self.visit_ref_helper(ctx.name())

        return KeyOptItem.from_kv(block_name, block_obj)

    def _do_import(
        self, ctx: ParserRuleContext, from_file: str, what_refs: list[str]
    ) -> KeyOptMap:
        """Return the import objects created by import statements."""

        _errors = []

        if not from_file:
            _errors.append(
                errors.AtoError("Expected a 'from <file-path>' after 'import'")
            )

        # get the current working directory
        current_file, _, _ = get_src_info_from_ctx(ctx)
        current_file = Path(current_file)
        if current_file.is_file():
            search_paths = chain((current_file.parent,), self.get_search_paths())
        else:
            search_paths = self.get_search_paths()

        for search_path in search_paths:
            candidate_path: Path = (search_path / from_file).resolve().absolute()
            if candidate_path.exists():
                break
        else:
            raise errors.AtoImportNotFoundError.from_ctx(  # pylint: disable=raise-missing-from
                ctx, f"File '{from_file}' not found."
            )

        imports = {}
        for _what_ref in what_refs:
            if not _what_ref:
                _errors.append(
                    errors.AtoError(
                        "Expected a name or attribute to import after 'import'"
                    )
                )

            if _what_ref == "*":
                # import everything
                raise NotImplementedError("import *")

            import_addr = address.add_entries(str(candidate_path), _what_ref)

            imports[_what_ref] = Import(
                src_ctx=ctx,
                obj_addr=import_addr,
            )

        return KeyOptMap(KeyOptItem.from_kv(k, v) for k, v in imports.items())

    def visitImport_stmt(self, ctx: ap.Import_stmtContext) -> KeyOptMap:
        """
        Updated import statement: from "abcd.ato" import xyz
        """
        from_file: str = self.visitString(ctx.string())
        what_refs = [
            self.visit_ref_helper(name_of_attr) for name_of_attr in ctx.name_or_attr()
        ]
        return self._do_import(ctx, from_file, what_refs)

    def visitDep_import_stmt(self, ctx: ap.Dep_import_stmtContext) -> KeyOptMap:
        """
        DEPRECATED: import xyz from "abcd.ato"
        """
        from_file: str = self.visitString(ctx.string())
        import_what_ref = self.visit_ref_helper(ctx.name_or_attr())
        return self._do_import(ctx, from_file, [import_what_ref])

    def visitBlocktype(self, ctx: ap.BlocktypeContext) -> Ref:
        """Return the address of a block type."""
        block_type_name = ctx.getText()
        match block_type_name:
            case "module":
                return Ref.from_one("MODULE")
            case "component":
                return Ref.from_one("COMPONENT")
            case "interface":
                return Ref.from_one("INTERFACE")
            case _:
                raise errors.AtoError(f"Unknown block type '{block_type_name}'")

    def visitRetype_stmt(self, ctx: ap.Retype_stmtContext) -> KeyOptMap:
        """TODO:"""
        # TODO: we should check the validity of the replacement here

        to_replace = self.visit_ref_helper(ctx.name_or_attr(0))
        new_class = self.visit_ref_helper(ctx.name_or_attr(1))

        replacement = Replacement(
            src_ctx=ctx,
            new_super_ref=new_class,
        )

        return KeyOptMap.from_kv(to_replace, replacement)

    def visitSimple_stmt(
        self, ctx: ap.Simple_stmtContext
    ) -> Iterable[_Sentinel | KeyOptItem]:
        """We have to be selective here to deal with the ignored children properly."""
        if ctx.retype_stmt() or ctx.import_stmt() or ctx.dep_import_stmt():
            return super().visitSimple_stmt(ctx)

        return KeyOptMap.empty()


def lookup_class_in_closure(context: ClassDef, ref: Ref) -> AddrStr:
    """
    This method finds an object in the closure of another object, traversing import statements.
    """
    assert context.closure is not None
    for scope in context.closure:
        obj_lead = scope.local_defs.get(ref[:1])
        import_leads = {
            imp_ref: imp
            for imp_ref, imp in scope.imports.items()
            if ref[0] == imp_ref[0]
        }

        if import_leads and obj_lead:
            # TODO: improve error message with details about what items are conflicting
            raise errors.AtoAmbiguousReferenceError.from_ctx(
                scope.src_ctx, f"Name '{ref[0]}' is ambiguous in '{scope}'."
            )

        if obj_lead is not None:
            if len(ref) > 1:
                raise NotImplementedError
            return obj_lead.address

        if ref in scope.imports:
            return scope.imports[ref].obj_addr

    if ref in BUILTINS_BY_REF:
        return BUILTINS_BY_REF[ref].address

    raise errors.AtoKeyError.from_ctx(
        context.src_ctx, f"Couldn't find {ref} in the scope of {context}"
    )


class BlockNotFoundError(errors.AtoKeyError):
    """
    Raised when a block doesn't exist.
    """


class Dizzy(HandleStmtsFunctional, HandlesPrimaries):
    """
    Dizzy is responsible for creating object layers, mixing cement,
    sand, aggregate, and water to create concrete.
    Ref.: https://www.youtube.com/watch?v=drBge9JyloA
    """

    def __init__(
        self,
        obj_def_getter: Callable[[AddrStr], ClassDef],
    ) -> None:
        self.obj_def_getter = obj_def_getter
        self._output_cache: dict[AddrStr, ClassLayer] = {
            k: v for k, v in BUILTINS_BY_ADDR.items()
        }
        self.class_def_scope: StackList[ClassDef] = StackList()
        super().__init__()

    def get_layer(self, addr: AddrStr) -> ClassLayer:
        """Returns the ObjectLayer for a given address."""
        if addr not in self._output_cache:
            obj_def = self.obj_def_getter(addr)
            obj = self.build_layer(obj_def)
            assert isinstance(obj, ClassLayer)
            self._output_cache[addr] = obj
        try:
            return self._output_cache[addr]
        except KeyError as ex:
            raise BlockNotFoundError(
                f"No block named $addr in {address.get_file(addr)}", addr=addr
            ) from ex

    def _get_supers_layer(self, cls_def: ClassDef) -> Optional[ClassLayer]:
        """Return the super object of a given object."""
        if cls_def.super_ref is not None:
            super_addr = lookup_class_in_closure(cls_def, cls_def.super_ref)
            return self.get_layer(super_addr)
        return None

    def build_layer(self, cls_def: ClassDef) -> ClassLayer:
        """Create an object layer from an object definition."""
        ctx = cls_def.src_ctx
        assert isinstance(ctx, (ap.File_inputContext, ap.BlockdefContext))

        # NOTE: visiting the block here relies upon the fact that both
        # file inputs and blocks have stmt children to be handled the same way.
        if isinstance(ctx, ap.BlockdefContext):
            ctx_with_stmts = ctx.block()
        else:
            ctx_with_stmts = ctx

        with self.class_def_scope.enter(cls_def):
            locals_ = self.visitBlock(ctx_with_stmts)

        strainer = locals_.strain()

        obj = ClassLayer(
            src_ctx=ctx_with_stmts,  # here we save something that's "block-like"
            obj_def=cls_def,
            super=self._get_supers_layer(cls_def),
            assignments={
                ref[0]: v for ref, v in strainer.strain(lambda x: isinstance(x.value, Assignment))
            }
        )

        if strainer:
            raise RuntimeError(f"Unexpected items in ClassLayer locals: {', '.join(strainer)}")

        return obj

    def visitFile_input(self, ctx: ap.File_inputContext) -> None:
        """I'm not sure how we'd end up here, but if we do, don't go down here"""
        raise RuntimeError("File inputs should not be visited")

    def visitBlockdef(self, ctx: ap.BlockdefContext) -> _Sentinel:
        """Don't go down blockdefs, they're just for defining objects."""
        return NOTHING

    def _get_type_info(self, ctx: ap.Declaration_stmtContext | ap.Assign_stmtContext) -> Optional[ClassLayer | pint.Unit]:
        """Return the type information from a type_info context."""
        if type_ctx := ctx.type_info():
            return type_ctx.name_or_attr().getText()
        return None

        # TODO: parse types properly
        if type_info := ctx.type_info():
            assert isinstance(type_info, ap.Type_infoContext)
            type_info_str: str = type_info.name_or_attr().getText()

            try:
                return lookup_class_in_closure(
                    self.class_def_scope.top,
                    type_info_str,
                )
            except KeyError:
                pass

            # TODO: implement types for ints, floats, strings, etc.
            # voltages, currents, lengths etc...

        return None

    def visitAssign_stmt(self, ctx: ap.Assign_stmtContext) -> KeyOptMap:
        assignable_ctx = ctx.assignable()
        assert isinstance(assignable_ctx, ap.AssignableContext)
        if assignable_ctx.new_stmt():
            # ignore new statements here, we'll deal with them in future layers
            return KeyOptMap.empty()

        assigned_value_ref = self.visitName_or_attr(ctx.name_or_attr())
        if len(assigned_value_ref) > 1:
            # we'll deal with overrides later too!
            return KeyOptMap.empty()

        assignment = Assignment(
            src_ctx=ctx,
            name=assigned_value_ref[0],
            value=self.visitAssignable(assignable_ctx),
            given_type=self._get_type_info(ctx),
        )
        return KeyOptMap.from_kv(assigned_value_ref, assignment)

    def visitDeclaration_stmt(self, ctx: ap.Declaration_stmtContext) -> KeyOptMap:
        """Handle declaration statements."""
        assigned_value_ref = self.visitName_or_attr(ctx.name_or_attr())
        if len(assigned_value_ref) > 1:
            raise errors.AtoSyntaxError(
                f"Can't declare fields in a nested object {assigned_value_ref}"
            )

        assignment = Assignment(
            src_ctx=ctx,
            name=assigned_value_ref[0],
            value=None,
            given_type=self._get_type_info(ctx),
        )
        return KeyOptMap.from_kv(assigned_value_ref, assignment)

    def visitSimple_stmt(
        self, ctx: ap.Simple_stmtContext
    ) -> Iterable[_Sentinel | KeyOptItem]:
        """We have to be selective here to deal with the ignored children properly."""
        if ctx.assign_stmt() or ctx.declaration_stmt():
            return super().visitSimple_stmt(ctx)

        return (NOTHING,)


@contextmanager
def _translate_addr_key_errors(ctx: ParserRuleContext):
    try:
        yield
    except KeyError as ex:
        addr = ex.args[0]
        terse_addr = address.get_instance_section(addr)
        raise errors.AtoKeyError.from_ctx(ctx, f"Couldn't find {terse_addr}") from ex


class Lofty(HandleStmtsFunctional, HandlesPrimaries):
    """Lofty's job is to walk orthogonally down (or really up) the instance tree."""

    def __init__(
        self,
        obj_layer_getter: Callable[[AddrStr], ClassLayer],
    ) -> None:
        self._output_cache: dict[AddrStr, Instance] = {}
        # known replacements are represented as the reference of the instance
        # to be replaced, and a tuple containing the length of the ref of the
        # thing that called for that replacement, and the object that will replace it
        self._known_replacements: dict[AddrStr, AddrStr] = {}
        self.obj_layer_getter = obj_layer_getter

        self._instance_addr_stack: StackList[AddrStr] = StackList()
        self._class_addr_stack: StackList[AddrStr] = StackList()
        super().__init__()

    @property
    def _current_instance(self) -> Instance:
        """Return the current instance."""
        return self._output_cache[self._instance_addr_stack.top]

    def get_instance(self, addr: AddrStr) -> Instance:
        """Return an instance object represented by the given address."""
        if addr in self._output_cache:
            return self._output_cache[addr]

        if address.get_instance_section(addr):
            # Trigger build of the tree above the instance
            self.get_instance(address.get_entry(addr))

        obj_layer = self.obj_layer_getter(addr)
        self.build_instance(addr, obj_layer)
        assert isinstance(self._output_cache[addr], Instance)

        return self._output_cache[addr]

    @contextmanager
    def apply_replacements_from_objs(
        self, objs: Iterable[ClassLayer]
    ) -> Iterable[AddrStr]:
        """
        Apply the replacements defined in the given objects,
        returning which replacements were applied
        """
        commanded_replacements = []

        for obj in objs:
            for ref, replacement in obj.obj_def.replacements.items():
                to_be_replaced_addr = address.add_instances(
                    self._instance_addr_stack.top, ref
                )
                if to_be_replaced_addr not in self._known_replacements:
                    replace_with_addr = lookup_class_in_closure(
                        obj.obj_def,
                        replacement.new_super_ref,
                    )

                    self._known_replacements[to_be_replaced_addr] = replace_with_addr
                    commanded_replacements.append(to_be_replaced_addr)
        try:
            yield
        finally:
            for ref in commanded_replacements:
                self._known_replacements.pop(ref)

    def build_instance(self, new_addr: AddrStr, super_obj: ClassLayer, src_ctx: Optional[ParserRuleContext] = None) -> None:
        """Create an instance from a reference and a super object layer."""

        if self._instance_addr_stack:
            # eg. we're not to the root
            parent_instance = self._output_cache[self._instance_addr_stack.top]
        else:
            # eg. we're at the root
            parent_instance = None


        new_instance = Instance.from_super(
            new_addr,
            super_obj,
            parent=parent_instance,
            src_ctx=src_ctx
        )
        self._output_cache[new_addr] = new_instance

        if self._instance_addr_stack:
            child_addr = address.get_name(new_addr)
            parent_instance.children[child_addr] = new_instance

        try:
            with ExitStack() as stack:
                stack.enter_context(self._instance_addr_stack.enter(new_addr))
                stack.enter_context(self.apply_replacements_from_objs(new_instance.supers))
                for super_obj_ in reversed(new_instance.supers):
                    stack.enter_context(self._class_addr_stack.enter(super_obj_.address))
                    if super_obj_.src_ctx is None:
                        # FIXME: this is currently the case for the builtins
                        continue

                    # visit the internals (eg. all the new statements, overrides etc...)
                    # of the things we're inheriting from
                    self.visitBlock(super_obj_.src_ctx)
        except Exception:
            if new_addr in self._output_cache:
                del self._output_cache[new_addr]
            raise

    def visitBlockdef(self, ctx: ap.BlockdefContext) -> _Sentinel:
        """Don't go down blockdefs, they're just for defining objects."""
        return NOTHING

    def handle_new_assignment(self, ctx: ap.Assign_stmtContext) -> KeyOptMap:
        """Specifically handle "something = new XYZ" assignments."""
        assigned_ref = self.visitName_or_attr(ctx.name_or_attr())

        assigned_name: str = assigned_ref[-1]
        assignable_ctx = ctx.assignable()
        assert isinstance(assignable_ctx, ap.AssignableContext)

        # FIXME: this is a giant fucking mess
        new_stmt = assignable_ctx.new_stmt()
        assert isinstance(new_stmt, ap.New_stmtContext)
        if len(assigned_ref) != 1:
            raise errors.AtoError(
                "Cannot assign a new object to a multi-part reference"
            )

        new_class_ref = self.visitName_or_attr(new_stmt.name_or_attr())

        new_addr = address.add_instance(
            self._instance_addr_stack.top, assigned_name
        )

        # Figure out what class to create the new instance from
        if new_addr in self._known_replacements:
            actual_super = self.obj_layer_getter(self._known_replacements[new_addr])
        else:
            try:
                current_obj_def = self.obj_layer_getter(
                    self._class_addr_stack.top
                ).obj_def
                new_class_addr = lookup_class_in_closure(
                    current_obj_def, new_class_ref
                )
            except KeyError as ex:
                raise errors.AtoKeyError.from_ctx(
                    current_obj_def.src_ctx, f"Couldn't find ref {new_class_ref}"
                ) from ex

            try:
                actual_super = self.obj_layer_getter(new_class_addr)
            except (BlockNotFoundError, errors.AtoFileNotFoundError) as ex:
                ex.set_src_from_ctx(ctx)
                raise ex

        # Create and register the new instance to the output cache
        self.build_instance(new_addr, actual_super, ctx)
        return KeyOptMap.empty()

    def visitAssign_stmt(self, ctx: ap.Assign_stmtContext) -> KeyOptMap:
        """Assignments override values and create new instance of things."""
        assigned_ref = self.visitName_or_attr(ctx.name_or_attr())

        assigned_name: str = assigned_ref[-1]
        assignable_ctx = ctx.assignable()
        assert isinstance(assignable_ctx, ap.AssignableContext)

        ########## Handle New Statements ##########
        if assignable_ctx.new_stmt():
            return self.handle_new_assignment(ctx)

        ########## Handle Overrides ##########

        # We've already dealt with direct assignments in the previous layer
        if len(assigned_ref) == 1:
            return KeyOptMap.empty()

        instance_addr_assigned_to = address.add_instances(
            self._instance_addr_stack.top, assigned_ref[:-1]
        )
        with _translate_addr_key_errors(ctx):
            instance_assigned_to = self._output_cache[instance_addr_assigned_to]

        # TODO: de-triplicate this
        if type_info := ctx.type_info():
            given_type = lookup_class_in_closure(
                self.obj_layer_getter(self._class_addr_stack.top).obj_def,
                type_info.name_or_attr()
            )
        else:
            given_type = None

        assignment = Assignment(
            src_ctx=ctx,
            name=assigned_name,
            value=self.visitAssignable(assignable_ctx),
            given_type=given_type,
        )

        instance_assigned_to.assignments[assigned_name].appendleft(assignment)

        return KeyOptMap.empty()

    def visit_pin_or_signal_helper(
        self, ctx: ap.Pindef_stmtContext | ap.Signaldef_stmtContext
    ) -> AddrStr:
        """This function makes a pin or signal instance and sticks it in the instance tree."""
        # NOTE: name has to come first because both have names,
        # but only pins have a "totally an integer"
        name = (ctx.name() or ctx.totally_an_integer()).getText()

        current_instance_addr = self._instance_addr_stack.top
        current_instance = self._output_cache[current_instance_addr]
        new_addr = address.add_instance(current_instance_addr, name)

        super_ = PIN if isinstance(ctx, ap.Pindef_stmtContext) else SIGNAL

        pin_or_signal = Instance.from_super(
            src_ctx=ctx,
            addr=new_addr,
            super_=super_,
            parent=current_instance,
        )

        self._output_cache[new_addr] = current_instance.children[name] = pin_or_signal

        return new_addr

    def visitPindef_stmt(self, ctx: ap.Pindef_stmtContext) -> KeyOptMap:
        """TODO:"""
        return self.visit_pin_or_signal_helper(ctx)

    def visitSignaldef_stmt(self, ctx: ap.Signaldef_stmtContext) -> KeyOptMap:
        """TODO:"""
        return self.visit_pin_or_signal_helper(ctx)

    def visitConnect_stmt(self, ctx: ap.Connect_stmtContext) -> KeyOptMap:
        """
        Connect interfaces together
        """
        source_addr = self.visitConnectable(ctx.connectable(0))
        target_addr = self.visitConnectable(ctx.connectable(1))

        with _translate_addr_key_errors(ctx):
            source_instance = self._output_cache[source_addr]
            target_instance = self._output_cache[target_addr]

        link = Link(
            src_ctx=ctx,
            parent=self._current_instance,
            source=source_instance,
            target=target_instance,
        )

        self._current_instance.links.append(link)

        return KeyOptMap.empty()

    def visitConnectable(self, ctx: ap.ConnectableContext) -> AddrStr:
        """TODO:"""
        if ctx.name_or_attr() or ctx.numerical_pin_ref():
            ref = self.visit_ref_helper(ctx.name_or_attr() or ctx.numerical_pin_ref())
            return address.add_instances(self._instance_addr_stack.top, ref)
        elif ctx.pindef_stmt() or ctx.signaldef_stmt():
            return self.visitChildren(ctx)
        else:
            raise ValueError("Unexpected context in visitConnectable")

    def visitSimple_stmt(self, ctx: ap.Simple_stmtContext) -> KeyOptMap:
        """We have to be selective here to deal with the ignored children properly."""
        if ctx.assign_stmt() or ctx.connect_stmt() or ctx.assert_stmt():
            return super().visitSimple_stmt(ctx)

        elif ctx.pindef_stmt() or ctx.signaldef_stmt():
            self.visitChildren(ctx)

        return KeyOptMap.empty()

    def visitAssert_stmt(self, ctx: ap.Assert_stmtContext) -> KeyOptMap:
        """Handle assertion statements."""
        comparison_ctx: ap.ComparisonContext = ctx.comparison()
        roley = Roley(self._instance_addr_stack.top)

        expressions_ = []
        operators = []

        def _add_expr_from_context(ctx: ap.Arithmetic_expressionContext):
            expr = Expression.from_numericish(
                roley.visit(ctx)
            )
            # TODO: this shouldn't be attached to the expression like this
            # as the only means to pretty-print them
            expr.src_ctx = ctx
            expressions_.append(expr)

        _add_expr_from_context(comparison_ctx.arithmetic_expression())

        for comp_ctx in comparison_ctx.compare_op_pair():
            assert isinstance(comp_ctx, ap.Compare_op_pairContext)
            if child_ctx := comp_ctx.lt_arithmetic_or():
                operators.append("<")
            elif child_ctx := comp_ctx.gt_arithmetic_or():
                operators.append(">")
            elif child_ctx := comp_ctx.in_arithmetic_or():
                operators.append("within")
            else:
                raise ValueError
            _add_expr_from_context(child_ctx.arithmetic_expression())

        assert len(expressions_) == len(operators) + 1

        assertions_ = [Assertion(
            src_ctx=ctx,
            lhs=expressions_[i],
            operator=operators[i],
            rhs=expressions_[i+1],
        ) for i in range(len(operators))]

        self._current_instance.assertions.extend(assertions_)

        return KeyOptMap.empty()


scoop = Scoop(parser.get_ast_from_file)
dizzy = Dizzy(scoop.get_obj_def)
lofty = Lofty(dizzy.get_layer)



import pint

from atopile.expressions import RangedValue


def test_comparitors():
    a = RangedValue(1, 2)
    b = RangedValue(2, 3)
    c = RangedValue(1, 3)

    assert a < b
    assert b > a
    assert not a.within(b)
    assert a.within(c)
    assert a == RangedValue(1, 2)


def test_units():
    a = RangedValue(1000, 2000, "mV")
    b = RangedValue(1, 2, "kV")
    assert a < b

    c = a * b
    assert c.unit == pint.Unit("V") ** 2


def test_arithmetic():
    assert RangedValue(1, 2) + RangedValue(2, 3) == RangedValue(3, 5)
    assert RangedValue(1, 2) - RangedValue(2, 3) == RangedValue(-2, 0)
    assert RangedValue(1, 2) * RangedValue(2, 3) == RangedValue(2, 6)
    assert RangedValue(1, 2) / RangedValue(2, 3) == RangedValue(1 / 3, 1)
    assert RangedValue(1, 2) ** 2 == RangedValue(1, 4)

    assert RangedValue(1, 2) + 2 == RangedValue(3, 4)
    assert RangedValue(1, 2) - 3 == RangedValue(-2, -1)
    assert RangedValue(1, 2) * 4 == RangedValue(4, 8)
    assert RangedValue(1, 2) / 5 == RangedValue(1 / 5, 2 / 5)

    assert 2 + RangedValue(1, 2) == RangedValue(3, 4)
    assert 3 - RangedValue(1, 2) == RangedValue(-2, -1)
    assert 4 * RangedValue(1, 2) == RangedValue(4, 8)
    assert 5 / RangedValue(1, 2) == RangedValue(5 / 2, 5)



# pylint: disable=too-few-public-methods

"""Schema and utils for atopile config files."""

import collections.abc
import fnmatch
import logging
from pathlib import Path
from typing import Any, Optional

import yaml
from attrs import Factory, define
from omegaconf import MISSING, OmegaConf
from omegaconf.errors import ConfigKeyError

import atopile.errors
from atopile import address

log = logging.getLogger(__name__)


CONFIG_FILENAME = "ato.yaml"
ATO_DIR_NAME = ".ato"
MODULE_DIR_NAME = "modules"
BUILD_DIR_NAME = "build"


@define
class ProjectPaths:
    """Config grouping for all the paths in a project."""

    src: Path = "./"
    layout: Path = "elec/layout"


@define
class ProjectBuildConfig:
    """Config for a build."""

    entry: str = MISSING
    targets: list[str] = ["__default__"]


@define
class ProjectServicesConfig:
    """A config for services used by the project."""
    components: str = "https://atopile-component-server-atsuhzfd5a-uc.a.run.app/jlc"


@define
class ProjectConfig:
    """
    The config object for atopile.
    """

    location: Path = MISSING

    ato_version: str = "0.1.0"
    paths: ProjectPaths = Factory(ProjectPaths)
    builds: dict[str, ProjectBuildConfig] = Factory(dict)
    dependencies: list[str] = []
    services: ProjectServicesConfig = Factory(ProjectServicesConfig)


KEY_CONVERSIONS = {
    "ato-version": "ato_version",
}


def _sanitise_key(key: str) -> str:
    """Sanitize a key."""
    return KEY_CONVERSIONS.get(key, key)


def _sanitise_item(item: tuple[Any, Any]) -> tuple[Any, Any]:
    """Sanitise the key of a dictionary item to be a valid python identifier."""
    k, v = item
    if isinstance(v, collections.abc.Mapping):
        return _sanitise_key(k), _sanitise_dict_keys(v)
    return _sanitise_key(k), v


def _sanitise_dict_keys(d: collections.abc.Mapping) -> collections.abc.Mapping:
    """Sanitise the keys of a dictionary to be valid python identifiers."""
    if d is None:
        return {}
    return dict(_sanitise_item(item) for item in d.items())


def make_config(project_config: Path) -> ProjectConfig:
    """
    Make a config object for a project.

    The typing on this is a little white lie... because they're really OmegaConf objects.
    """
    structure: ProjectConfig = OmegaConf.structured(ProjectConfig())

    with project_config.open() as f:
        config_data = yaml.safe_load(f)
    project_config_data = OmegaConf.create(_sanitise_dict_keys(config_data))

    structure.location = project_config.parent.expanduser().resolve().absolute()

    for _ in range(1000):
        try:
            return OmegaConf.merge(
                structure,
                project_config_data,
            )
        except ConfigKeyError as ex:
            dot_path = ex.full_key.split(".")
            container = project_config_data
            for key in dot_path[:-1]:
                container = container[key]
            del container[dot_path[-1]]

            atopile.errors.AtoError(
                f"Unknown config option in {structure.location}. Ignoring \"{ex.full_key}\".",
                title="Unknown config option",
            ).log(log, logging.WARNING)
    raise atopile.errors.AtoError("Too many config errors")


def get_project_dir_from_path(path: Path) -> Path:
    """
    Resolve the project directory from the specified path.
    """
    #TODO: when provided with the "." path, it doesn't find the config in parent directories
    path = Path(path)
    for p in [path] + list(path.parents):
        clean_path = p.resolve().absolute()
        if (clean_path / CONFIG_FILENAME).exists():
            return clean_path
    raise atopile.errors.AtoFileNotFoundError(
        f"Could not find {CONFIG_FILENAME} in {path} or any parents"
    )


_loaded_configs: dict[Path, str] = {}


def get_project_config_from_path(path: Path) -> ProjectConfig:
    """
    Get the project config from a path.
    """
    project_dir = get_project_dir_from_path(path)
    project_config_file = project_dir / CONFIG_FILENAME
    if project_config_file not in _loaded_configs:
        _loaded_configs[project_config_file] = make_config(project_config_file)
    return _loaded_configs[project_config_file]


def get_project_config_from_addr(addr: str) -> ProjectConfig:
    """
    Get the project config from an address.
    """
    return get_project_config_from_path(Path(address.get_file(addr)))


# FIXME: we need factory constructors for these classes
@define
class ProjectContext:
    """A class to hold the arguments to a project."""

    project_path: Path  # abs path to the project directory
    src_path: Path  # abs path to the source directory
    module_path: Path  # abs path to the module directory
    layout_path: Path  # eg. path/to/project/layouts/default/default.kicad_pcb
    config: ProjectConfig

    @classmethod
    def from_config(cls, config: ProjectConfig) -> "ProjectContext":
        """Create a BuildArgs object from a Config object."""

        return ProjectContext(
            project_path=Path(config.location),
            src_path=Path(config.location) / config.paths.src,
            module_path=Path(config.location) / ATO_DIR_NAME / MODULE_DIR_NAME,
            layout_path=Path(config.location) / config.paths.layout,
            config=config,
        )

    @classmethod
    def from_path(cls, path: Path) -> "ProjectContext":
        """Create a BuildArgs object from a Config object."""
        return cls.from_config(get_project_config_from_path(path))


def match_user_layout(path: Path) -> bool:
    """Check whether a given filename is a KiCAD autosaved layout."""
    autosave_patterns = [
        "_autosave-*",
        "*-save.kicad_pcb",
    ]

    name = path.name

    for pattern in autosave_patterns:
        if fnmatch.fnmatch(name, pattern):
            return False
    return True


def find_layout(layout_base: Path) -> Optional[Path]:
    """Return the layout associated with a build."""
    if layout_base.with_suffix(".kicad_pcb").exists():
        return layout_base.with_suffix(".kicad_pcb").resolve().absolute()
    elif layout_base.is_dir():
        layout_candidates = list(
            filter(
                match_user_layout,
                layout_base.glob("*.kicad_pcb")
            )
        )

        if len(layout_candidates) == 1:
            return layout_candidates[0].resolve().absolute()

        else:
            raise atopile.errors.AtoError(
                "Layout directories must contain only 1 layout,"
                f" but {len(layout_candidates)} found in {layout_base}"
            )


@define
class BuildContext:
    """A class to hold the arguments to a build."""
    project_context: ProjectContext

    name: str

    entry: address.AddrStr  # eg. "path/to/project/src/entry-name.ato:module.path"
    targets: list[str]

    layout_path: Optional[Path]  # eg. path/to/project/layouts/default/default.kicad_pcb
    build_path: Path  # eg. path/to/project/build/<build-name>

    output_base: Path  # eg. path/to/project/build/<build-name>/entry-name

    @classmethod
    def from_config(
        cls,
        config_name: str,
        build_config: ProjectBuildConfig,
        project_context: ProjectContext
    ) -> "BuildContext":
        """Create a BuildArgs object from a Config object."""
        abs_entry = address.AddrStr(project_context.project_path / build_config.entry)

        build_path = Path(project_context.project_path) / BUILD_DIR_NAME

        return BuildContext(
            project_context=project_context,
            name=config_name,
            entry=abs_entry,
            targets=build_config.targets,
            layout_path=find_layout(project_context.project_path / project_context.layout_path / config_name),
            build_path=build_path,
            output_base=build_path / config_name,
        )

    @classmethod
    def from_config_name(cls, config: ProjectConfig, build_name: str) -> "BuildContext":
        """Create a BuildArgs object from a Config object."""
        project_context = ProjectContext.from_config(config)

        try:
            build_config = config.builds[build_name]
        except KeyError as ex:
            raise atopile.errors.AtoError(
                f"Build {build_name} not found for project {config.location}\n"
                f"Available builds: {list(config.builds.keys())}"
            ) from ex

        return cls.from_config(build_name, build_config, project_context)


_project_context: Optional[ProjectContext] = None


def set_project_context(project_context: ProjectContext) -> None:
    """
    Set the project context for the current process.
    """
    global _project_context
    _project_context = project_context


def get_project_context() -> ProjectContext:
    """
    Get the project context for the current process.
    """
    if _project_context is None:
        raise ValueError("Project context not set")
    return _project_context



"""CLI command definition for `ato build`."""
import json
import logging
import shutil
from functools import wraps
from typing import Callable, Optional

import click

import atopile.bom
import atopile.config
import atopile.front_end
import atopile.layout
import atopile.manufacturing_data
import atopile.netlist
import atopile.assertions
from atopile.cli.common import project_options
from atopile.components import configure_cache, download_footprint
from atopile.config import BuildContext
from atopile.errors import handle_ato_errors, iter_through_errors
from atopile.instance_methods import all_descendants, match_components
from atopile.netlist import get_netlist_as_str

log = logging.getLogger(__name__)


@click.command()
@project_options
def build(build_ctxs: list[BuildContext]):
    """
    Build the specified --target(s) or the targets specified by the build config.
    Specify the root source file with the argument SOURCE.
    eg. `ato build --target my_target path/to/source.ato:module.path`
    """
    for err_cltr, build_ctx in iter_through_errors(build_ctxs):
        log.info("Building %s", build_ctx.name)
        with err_cltr():
            _do_build(build_ctx)

    project_context = atopile.config.get_project_context()

    # FIXME: this should be done elsewhere, but there's no other "overview"
    # that can see all the builds simultaneously
    manifest = {}
    manifest["version"] = "2.0"
    for ctx in build_ctxs:
        if ctx.layout_path:
            by_layout_manifest = manifest.setdefault("by-layout", {}).setdefault(str(ctx.layout_path), {})
            by_layout_manifest["layouts"] = str(ctx.output_base.with_suffix(".layouts.json"))

    manifest_path = project_context.project_path / "build" / "manifest.json"
    manifest_path.parent.mkdir(exist_ok=True, parents=True)
    with open(manifest_path, "w", encoding="utf-8") as f:
        json.dump(manifest, f)


def _do_build(build_ctx: BuildContext) -> None:
    """Execute a specific build."""

    # Configure the cache for component data
    # TODO: flip this around so that the cache pulls what it needs from the
    # project context
    configure_cache(atopile.config.get_project_context().project_path)

    # Solve the unknown variables
    atopile.assertions.solve_assertions(build_ctx)

    # Ensure the build directory exists
    log.info("Writing outputs to %s", build_ctx.build_path)
    build_ctx.build_path.mkdir(parents=True, exist_ok=True)

    if build_ctx.targets == ["__default__"]:
        targets = muster.do_by_default
    elif build_ctx.targets == ["*"] or build_ctx.targets == ["all"]:
        targets = list(muster.targets.keys())
    else:
        targets = build_ctx.targets

    build_ctx.output_base.parent.mkdir(parents=True, exist_ok=True)

    for err_cltr, target_name in iter_through_errors(targets):
        log.info("Building %s", target_name)
        with err_cltr():
            muster.targets[target_name](build_ctx)


TargetType = Callable[[BuildContext], None]


class Muster:
    """A class to register targets to."""

    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.targets = {}
        self.do_by_default = []
        self.log = logger or logging.getLogger(__name__)

    def add_target(self, func: TargetType, name: Optional[str] = None, default: bool = True):
        """Register a function as a target."""
        name = name or func.__name__
        self.targets[name] = func
        if default:
            self.do_by_default.append(name)
        return func

    def register(self, name: Optional[str] = None, default: bool = True):
        """Register a target under a given name."""

        def decorator(func: TargetType):
            @wraps(func)
            def wrapper(build_args: BuildContext):
                with handle_ato_errors():
                    return func(build_args)

            self.add_target(wrapper, name, default)
            return wrapper

        return decorator


muster = Muster()


@muster.register("copy-footprints")
def consolidate_footprints(build_args: BuildContext) -> None:
    """Consolidate all the project's footprints into a single directory."""
    fp_target = build_args.build_path / "footprints" / "footprints.pretty"
    fp_target.mkdir(exist_ok=True, parents=True)

    for fp in atopile.config.get_project_context().project_path.glob("**/*.kicad_mod"):
        try:
            shutil.copy(fp, fp_target)
        except shutil.SameFileError:
            log.debug("Footprint %s already exists in the target directory", fp)


@muster.register("netlist")
def generate_netlist(build_args: BuildContext) -> None:
    """Generate a netlist for the project."""
    with open(build_args.output_base.with_suffix(".net"), "w", encoding="utf-8") as f:
        f.write(get_netlist_as_str(build_args.entry))


@muster.register("bom")
def generate_bom(build_args: BuildContext) -> None:
    """Generate a BOM for the project."""
    with open(build_args.output_base.with_suffix(".csv"), "w", encoding="utf-8") as f:
        f.write(atopile.bom.generate_bom(build_args.entry))


@muster.register("designator-map")
def generate_designator_map(build_args: BuildContext) -> None:
    """Generate a designator map for the project."""
    atopile.bom.generate_designator_map(build_args.entry)


@muster.register("mfg-data", default=False)
def generate_manufacturing_data(build_ctx: BuildContext) -> None:
    """Generate a designator map for the project."""
    atopile.manufacturing_data.generate_manufacturing_data(build_ctx)


@muster.register("clone-footprints")
def clone_footprints(build_args: BuildContext) -> None:
    """Clone the footprints for the project."""
    all_components = filter(match_components, all_descendants(build_args.entry))

    for component in all_components:
        log.debug("Cloning footprint for %s", component)
        download_footprint(component, footprint_dir=build_args.build_path / "footprints/footprints.pretty")


@muster.register("layout-module-map")
def generate_module_map(build_args: BuildContext) -> None:
    """Generate a designator map for the project."""
    atopile.layout.generate_module_map(build_args)


@muster.register("assertions-report")
def generate_assertion_report(build_ctx: BuildContext) -> None:
    """Generate a report based on assertions made in the source code."""
    atopile.assertions.generate_assertion_report(build_ctx)



"""
Generate a report based on assertions made in the source code.
"""

import itertools
import logging
import textwrap
from collections import ChainMap, defaultdict
from enum import Enum
from typing import Any, Iterable

import pint
import rich
from eseries import eseries
from rich.style import Style
from rich.table import Table
from scipy.optimize import minimize

from atopile import address, config, errors, instance_methods, loop_soup, telemetry
from atopile.front_end import (
    Assertion,
    Assignment,
    Expression,
    RangedValue,
    lofty,
)

log = logging.getLogger(__name__)

light_row = Style(color="bright_black")
dark_row = Style(color="white")


class AssertionStatus(Enum):
    """
    The status of an assertion.
    """

    PASSED = "[green]Passed[/]"
    FAILED = "[red]Failed[/]"
    ERROR = "[red]Error[/]"


def generate_assertion_report(build_ctx: config.BuildContext):
    """
    Generate a report based on assertions made in the source code.
    """
    table = Table(show_header=True, header_style="bold green")
    table.add_column("Status")
    table.add_column("Assertion")
    table.add_column("Numeric")
    table.add_column("Address")
    table.add_column("Notes")

    rows = 0

    def _add_row(
        status: AssertionStatus,
        assertion_str: str,
        numeric: str,
        addr: address.AddrStr,
        notes: str,
    ):
        nonlocal rows
        table.add_row(
            status.value,
            assertion_str,
            numeric,
            address.get_instance_section(addr),
            notes,
            style=dark_row if rows % 2 else light_row,
        )
        rows += 1

    context = {}
    for instance_addr in instance_methods.all_descendants(build_ctx.entry):
        instance = lofty.get_instance(instance_addr)
        for assertion in instance.assertions:
            try:
                telemetry.log_assertion(str(assertion))
            except Exception as e:
                log.debug("Failed to log assertion: %s", e)
            new_symbols = {
                s.key for s in assertion.lhs.symbols | assertion.rhs.symbols
            } - context.keys()
            for symbol in new_symbols:
                parent_inst = instance_methods.get_instance(
                    address.get_parent_instance_addr(symbol)
                )
                assign_name = address.get_name(symbol)

                if assign_name not in parent_inst.assignments:
                    raise errors.AtoKeyError(
                        f"No attribute '{assign_name}' bound on '{parent_inst.addr}'"
                    )

                assignment = parent_inst.assignments[assign_name][0]
                if assignment.value is None:
                    raise errors.AtoTypeError(
                        f"'{symbol}' is defined, but has no value"
                    )
                context[symbol] = assignment.value

            try:
                a = assertion.lhs(context)
                b = assertion.rhs(context)
            except errors.AtoError as e:
                _add_row(
                    AssertionStatus.ERROR, str(assertion), "", instance_addr, str(e)
                )
                raise

            if _do_op(a, assertion.operator, b):
                status = AssertionStatus.PASSED
            else:
                status = AssertionStatus.FAILED

            numeric = a.pretty_str() + " " + assertion.operator + " " + b.pretty_str()

            _add_row(status, str(assertion), numeric, instance_addr, "")

    rich.print(table)


def _do_op(a: RangedValue, op: str, b: RangedValue) -> bool:
    """Perform the operation specified by the operator."""
    if op == "within":
        return a.within(b)
    elif op == "<":
        return a < b
    elif op == ">":
        return a > b
    else:
        raise ValueError(f"Unrecognized operator: {op}")


def _check_assertion(assertion: Assertion, context: dict) -> bool:
    """
    Check if an assertion is true in the given context.
    """
    try:
        a = assertion.lhs(context)
        b = assertion.rhs(context)
        return _do_op(a, assertion.operator, b)
    except pint.errors.DimensionalityError as ex:
        raise errors.AtoTypeError(
            f"Dimensionality mismatch in assertion: {assertion}"
            f" ({ex.units1} incompatible with {ex.units2})"
        ) from ex


def solve_assertions(build_ctx: config.BuildContext):
    """
    Solve the assertions in the build context.
    FIXME:
        - This mutates the instances
    """

    # 1. Find all the symbols referenced in assertions in the design
    # ... and figure out which are variables and which are fixed
    referenced_symbols: set[address.AddrStr] = set()
    constants: dict[address.AddrStr, Any] = {}
    variable_units: dict[str, pint.Unit] = {}
    variable_soup = loop_soup.LoopSoup()
    for error_collector, instance_addr in errors.iter_through_errors(
        instance_methods.all_descendants(build_ctx.entry)
    ):
        instance = lofty.get_instance(instance_addr)
        for assertion in instance.assertions:
            with error_collector():
                # Bucket new symbols into variables and constants
                assertion_symbols = {
                    s.key for s in assertion.lhs.symbols | assertion.rhs.symbols
                }
                new_symbols = assertion_symbols - referenced_symbols
                referenced_symbols |= new_symbols
                for symbol in new_symbols:
                    parent_inst = instance_methods.get_instance(
                        address.get_parent_instance_addr(symbol)
                    )
                    assign_name = address.get_name(symbol)

                    if assign_name not in parent_inst.assignments:
                        raise errors.AtoKeyError(
                            f"No attribute '{assign_name}' bound on '{parent_inst.addr}'"
                        )

                    assignment = parent_inst.assignments[assign_name][0]
                    if assignment.value is None:
                        # TEMP: this is in place because our discretization strategy
                        # requires E96 things
                        if not assignment.unit.is_compatible_with(pint.Unit("ohm")):
                            raise errors.AtoTypeError.from_ctx(
                                assignment.src_ctx,
                                f"'{symbol}' is defined, but has no value.\n"
                                "Currently the calculator only supports resistor "
                                "values, so please assign a value to this attribute.",
                            )
                        variable_units[symbol] = assignment.unit
                        variable_soup.add(symbol)
                    else:
                        constants[symbol] = assignment.value

                # Group up all the entangled symbols
                variable_soup.join_multiple(
                    filter(lambda s: s in variable_soup, assertion_symbols)
                )

    # If there isn't anything to solve, just return
    if not variable_soup:
        log.info("No variables in assertions to solve")
        return

    vars = list(variable_soup)
    try:
        telemetry.log_eqn_vars(len(vars))
    except Exception as e:
        log.debug("Failed to log eqn_vars: %s", e)
    log.debug("Variables to solve for: %s", vars)

    # 2. Create groups of assertions based on the symbols they contain
    # This way we don't need to solve for every assertion at once
    assertion_groups: dict[tuple[address.AddrStr], list[dict]] = defaultdict(list)
    for instance_addr in instance_methods.all_descendants(build_ctx.entry):
        instance = lofty.get_instance(instance_addr)
        for assertion in instance.assertions:
            for symbol in assertion.lhs.symbols | assertion.rhs.symbols:
                # Only add the constraint if the assertion contains a variable
                # Otherwise, this assertion isn't relevant to the optimization
                if symbol.key in variable_soup:
                    group_key = tuple(
                        sorted(variable_soup.get_loop(symbol.key).iter_values())
                    )
                    assertion_groups[group_key].append(assertion)
                    break  # onto the next assertion

    # 3. Solve each group
    table = Table(show_header=True, header_style="bold green")
    table.add_column("Address")
    table.add_column("Value")
    row_count = 0

    for error_collector, assertion_group in errors.iter_through_errors(assertion_groups.items()):
        group_vars, assertions = assertion_group
        with error_collector():
            log.debug("Solving for group %s", group_vars)
            translator = _translator_factory(
                group_vars, [variable_units[addr] for addr in group_vars], constants
            )

            constraints = [
                c for a in assertions for c in _constraint_factory(a, translator)
            ]

            result = minimize(
                # FIXME: see notes in _cost about how stupid this function is
                _cost,
                # FIXME: this single starting points is medicore at best
                tuple(v + 10 for v in range(len(group_vars) * 2)),
                constraints=constraints,
                # FIXME: this should have bounds, but they'll depend on the variable's constraints
                # We could perhaps inherit these from the variable's type?
                bounds=[(0, None)] * len(group_vars) * 2,
                options={
                    "disp": log.getEffectiveLevel() <= logging.DEBUG,
                    "maxiter": 1000,
                },
            )
            log.debug("Optimization result: %s", result)
            if not result.success:
                title = f"Failed to solve assertions: {result.message}"
                msg = textwrap.dedent(
                    """
                    The optimization algorithm failed to find a solution.
                    This could mean a litany of things went wrong, but most likely:
                    - The constraints are backwards / too tight
                    - Variables are missing tolerances
                    - Assertions conflict with one another
                    """
                )

                if (
                    len(assertions) == 1
                    and hasattr(assertions[0], "src_ctx")
                    and assertions[0].src_ctx
                ):
                    raise errors.AtoError.from_ctx(
                        assertions[0].src_ctx,
                        title=title,
                        message=msg,
                    )
                else:
                    raise errors.AtoError(
                        msg,
                        title=title,
                    )

            # Here we're attempting to shuffle the values into eseries
            result_means = [
                (result.x[i * 2] + result.x[i * 2 + 1]) / 2 for i in range(len(group_vars))
            ]
            for r_vals in itertools.product(
                *[eseries.find_nearest_few(eseries.E96, x_val) for x_val in result_means],
                repeat=1,
            ):
                final_values = [
                    v
                    for r_val in r_vals
                    for v in [
                        r_val - 1.1 * r_val * eseries.tolerance(eseries.E96),
                        r_val + 1.1 * r_val * eseries.tolerance(eseries.E96),
                    ]
                ]
                _context = translator(final_values)
                if all(_check_assertion(a, _context) for a in assertions):
                    break
            else:
                raise errors.AtoError(
                    "Failed to find a solution that satisfies all the assertions using e96 values",
                    title="Failed to solve assertions",
                )

            # Apply the values back to the model
            for i, addr in enumerate(group_vars):
                parent = lofty.get_instance(address.get_parent_instance_addr(addr))
                name = address.get_name(addr)

                val = RangedValue(
                    final_values[i * 2], final_values[i * 2 + 1], variable_units[addr]
                )

                table.add_row(
                    address.get_instance_section(addr),
                    str(val),
                    style=dark_row if row_count % 2 else light_row,
                )
                row_count += 1

                # FIXME: Do we want to mutate the model here?
                # FIXME: Creating Assignment object here is annoying
                parent.assignments[name].appendleft(
                    Assignment(name, value=val, given_type="None")
                )

    # Solved for assertion values
    log.info("Values for solved variables:")
    rich.print(table)


def _translator_factory(
    args: Iterable, arg_units: Iterable[pint.Unit], known_context: dict
):
    assert len(args) == len(arg_units)

    def _translate(x):
        assert len(x) == len(args) * 2
        arg_ctx = {
            arg: RangedValue(x[i * 2], x[i * 2 + 1], arg_units[i])
            for i, arg in enumerate(args)
        }
        return ChainMap(arg_ctx, known_context)

    return _translate


def _tolerance_cost(min_: float, max_: float):
    if min_ == max_:
        return 1e4
    mean = (min_ + max_) / 2
    if mean == 0:
        return 1 / (100 * (max_ - min_)) ** 2
    return 1 / (100 * (max_ - min_) / mean) ** 2


def _cost(x):
    """
    Works under the assumption that x is paired up ranged values
    FIXME:
        - This assumption of tolerance having cost breaks for variables
        you'd expect to expand into one another, eg. v_in of a vdiv used.
        for feedback on a reg or i_q for it's quiescent current. This
        manifests as squeezing the resistors' tolerances down, which means
        we frequently can't find any that actually fit.
    """
    return sum(
        _tolerance_cost(x[i * 2], x[i * 2 + 1]) for i in range(len(x) // 2)
    ) / len(x)


def _constraint_factory(assertion: Assertion, translator):
    def lower_than(a: Expression, b: Expression) -> list[dict]:
        def _brrr(x):
            ctx = translator(x)
            return (
                b(ctx).min_qty.to_base_units().magnitude
                - a(ctx).max_qty.to_base_units().magnitude
            )

        return [
            {"type": "ineq", "fun": _brrr},
        ]

    def greater_than(a: Expression, b: Expression) -> list[dict]:
        def _brrr(x):
            ctx = translator(x)
            return (
                a(ctx).min_qty.to_base_units().magnitude
                - b(ctx).max_qty.to_base_units().magnitude
            )

        return [
            {"type": "ineq", "fun": _brrr},
        ]

    def within(a: Expression, b: Expression) -> list[dict]:
        def _brrr(x):
            ctx = translator(x)
            return (
                b(ctx).max_qty.to_base_units().magnitude
                - a(ctx).max_qty.to_base_units().magnitude
            )

        def _brrr2(x):
            ctx = translator(x)
            return (
                a(ctx).min_qty.to_base_units().magnitude
                - b(ctx).min_qty.to_base_units().magnitude
            )

        return [
            {"type": "ineq", "fun": _brrr},
            {"type": "ineq", "fun": _brrr2},
        ]

    if assertion.operator == "<":
        return lower_than(assertion.lhs, assertion.rhs)

    elif assertion.operator == ">":
        return greater_than(assertion.lhs, assertion.rhs)

    elif assertion.operator == "within":
        return within(assertion.lhs, assertion.rhs)

    raise ValueError(f"Unknown operator {assertion.operator}")



import pytest
from atopile.nets import _Net

@pytest.fixture
def net():
    return _Net([])

@pytest.mark.parametrize(
    "prefix, base, suffix, expected",
    [
        (None, None, None, "net"),
        ("a", None, None, "a-net"),
        (None, "b", None, "b"),
        (None, None, 3, "net-3"),
        ("a", "b", None, "a-b"),
        ("a", "b", 3, "a-b-3"),
    ]
)
def test_net_name(net: _Net, prefix, base, suffix, expected):
    net.prefix = prefix
    net.base_name = base
    net.suffix = suffix
    assert net.get_name() == expected



from antlr4 import InputStream

from atopile.expressions import RangedValue, Expression, Symbol
from atopile.front_end import Roley
from atopile.parse import make_parser
from atopile.parser.AtopileParser import AtopileParser


def parse(src_code: str) -> AtopileParser.Arithmetic_expressionContext:
    """Parse a string as a file input."""
    input = InputStream(src_code)
    input.name = "<inline>"
    parser = make_parser(input)
    return parser.arithmetic_expression()


roley = Roley("//a:b::")


def _run(src_code: str):
    return roley.visit(parse(src_code))


def test_basic_expression():
    assert _run("1 + 2") == 3
    assert _run("1 - 2") == -1
    assert _run("1 * 2") == 2
    assert _run("1 / 2") == 0.5
    assert _run("1 ** 2") == 1
    assert _run("1 + 2 * 3") == 7
    assert _run("1 * 2 + 3") == 5
    assert _run("1 * (2 + 3)") == 5
    assert _run("(1 * 2) ** 3") == 8


def test_units():
    assert _run("1 V + 205 mV") == RangedValue(1205, 1205, "mV")
    assert _run("1V + 205mV") == RangedValue(1205, 1205, "mV")
    assert _run("1V ± 205mV + 1V ± 1mV") == RangedValue(1794, 2206, "mV")


def test_parens():
    assert _run("(1 + 2) * 3") == 9
    assert _run("1 + (2 * 3)") == 7
    assert _run("1 * (2 + 3)") == 5
    assert _run("1 * (2 + 3) * 4") == 20
    assert _run("1 * (2 + 3) * 4 + 5") == 25
    assert _run("1 * (2 + 3) * (4 + 5)") == 45
    assert _run("1 * (2 + 3) * (4 + 5) + 6") == 51


def test_simple_pseudo_symbols():
    a = _run("1 + 2 + c.d.e")
    assert callable(a)
    assert isinstance(a, Expression)
    assert a({"//a:b::c.d.e": 5}) == 8
    assert a.symbols == {Symbol("//a:b::c.d.e")}


def test_pseudo_symbols():
    context = {
        "//a:b::a.a.a": 4,
        "//a:b::b": 3,
        "//a:b::c": 9,
    }
    assert _run("(a.a.a**b + 17) / c + 53")(context) == 62



"""
Datatypes used in the model.
"""

import logging
from contextlib import contextmanager
from typing import (
    Callable,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Type,
    TypeVar,
)

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class Ref(tuple[str]):
    """Shell class to provide basic utils for a reference."""

    def add_name(self, name: str | int) -> "Ref":
        """Return a new Ref with the given name."""
        return Ref((*self, name))

    def __str__(self) -> str:
        return ".".join(map(str, self))

    @classmethod
    def empty(cls) -> "Ref":
        """Return an empty Ref."""
        return cls(())

    @classmethod
    def from_one(cls, name: str | int) -> "Ref":
        """Return a Ref with a single item."""
        return cls((name,))


T = TypeVar("T")


class KeyOptItem(tuple[Optional[Ref], T], Generic[T]):
    """A class representing anf optionally-named thing."""

    @property
    def ref(self) -> Optional[Ref]:
        """Return the name of this item, if it has one."""
        return self[0]

    @property
    def value(self) -> T:
        """Return the value of this item."""
        return self[1]

    @classmethod
    def from_kv(cls, key: Optional[Ref], value: T) -> "KeyOptItem[T]":
        """Return a KeyOptItem with a single item."""
        return KeyOptItem((key, value))


class KeyOptMap(tuple[KeyOptItem[T]]):
    """A class representing a set of optionally-named things."""

    def named_items(self) -> Mapping[Ref, T]:
        """Return all the named items in this set, ignoring the unnamed ones."""
        return dict(filter(lambda x: x.ref is not None, self))

    def map_items_by_type(
        self, types: Iterable[Type | Iterable[Type]]
    ) -> Mapping[Type, "KeyOptMap[T]"]:
        """Return a mapping of items in this set by type."""
        return {t: tuple(self.filter_items_by_type(t)) for t in types}

    def unnamed_items(self) -> Iterable[T]:
        """Return an interable of all the unnamed items in this set."""
        return map(lambda x: x.value, filter(lambda x: x.ref is None, self))

    def filter_items_by_type(self, types: Type | Iterable[Type]) -> Iterator[T]:
        """Helper function to filter by type."""
        return filter(lambda x: isinstance(x.value, types), self)

    def keys(self) -> Iterable[Ref]:
        """Return an iterable of all the names in this set."""
        return map(lambda x: x.ref, filter(lambda x: x.ref is not None, self))

    def values(self) -> Iterable[T]:
        """Return an iterable of all the values in this set."""
        return map(lambda x: x.value, self)

    def strain(self) -> "Strainer[KeyOptItem[T]]":
        """Return a Strainer for this KeyOptMap."""
        return Strainer(self)

    @classmethod
    def from_item(cls, item: KeyOptItem[T]) -> "KeyOptMap[T]":
        """Return a KeyOptMap with a single item."""
        return KeyOptMap((item,))

    @classmethod
    def from_kv(cls, key: Optional[Ref], value: T) -> "KeyOptMap[T]":
        """Return a KeyOptMap with a single item."""
        return cls.from_item(KeyOptItem.from_kv(key, value))

    @classmethod
    def empty(cls) -> "KeyOptMap[T]":
        """Return an empty KeyOptMap."""
        return cls(())


class Strainer(list[T]):
    """A class to pop filtered things from a list."""

    def __init__(self, items: Iterable[T]) -> None:
        super().__init__(items)

    def iter_strain(self, filter_: Callable[[T], bool]) -> Iterator[T]:
        """Yield items that pass the filter, removing them from the list."""
        idx = 0
        while idx < len(self):
            item = self[idx]
            if filter_(item):
                yield self.pop(idx)
            else:
                idx += 1

    def strain(self, filter_: Callable[[T], bool]) -> list[T]:
        """Return a list of items that pass the filter, removing them from the list."""
        return list(self.iter_strain(filter_))


class StackList(list[T]):
    """Manages context while compiling ato code."""
    def __init__(self) -> None:
        super().__init__([])

    @contextmanager
    def enter(self, context: T) -> None:
        """Enter a context."""
        self.append(context)
        try:
            yield
        finally:
            self.pop()

    @property
    def top(self) -> T:
        """Return the current context."""
        return self[-1]


class DotDict(dict):
    """A dict you can dot"""

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as ex:
            raise AttributeError(f"Attribute {name} not found", name) from ex



"""
This module contains functions for interacting with layout data,
and generating files required to reuse layouts.

Thanks @nickkrstevski (https://github.com/nickkrstevski) for
the heavy lifting on this one!
"""

import hashlib
import json
import logging
import uuid
from collections import defaultdict

from atopile import address, config, errors
from atopile.instance_methods import (
    all_descendants,
    find_matching_super,
    match_components,
    match_modules,
)

log = logging.getLogger(__name__)


def generate_uuid_from_string(path: str) -> str:
    """Spits out a uuid in hex from a string"""
    path_as_bytes = path.encode("utf-8")
    hashed_path = hashlib.blake2b(path_as_bytes, digest_size=16).digest()
    return str(uuid.UUID(bytes=hashed_path))


def generate_comp_uid(comp_addr: str) -> str:
    """Get a unique identifier for a component."""
    instance_section = address.get_instance_section(comp_addr)
    if not instance_section:
        raise ValueError(f"Component address {comp_addr} has no instance section")
    return generate_uuid_from_string(instance_section)


def _find_module_layouts() -> dict[str, list[config.BuildContext]]:
    """
    Return a dict of all the known entry points of dependencies in the project.
    The dict maps the entry point's address to another map of the entry point's
    build name and the layout file path.
    """
    directory = config.get_project_context().project_path

    entries = defaultdict(list)
    for filepath in directory.glob("**/ato.yaml"):
        cfg = config.get_project_config_from_path(filepath)

        for build_name in cfg.builds:
            ctx = config.BuildContext.from_config_name(cfg, build_name)
            entries[ctx.entry].append(ctx)

    return entries


def generate_module_map(build_ctx: config.BuildContext) -> None:
    """Generate a file containing a list of all the modules and their components in the build."""
    module_map = {}

    laid_out_modules = _find_module_layouts()
    for module_instance in filter(match_modules, all_descendants(build_ctx.entry)):
        module_super = find_matching_super(module_instance, list(laid_out_modules.keys()))
        if not module_super:
            continue

        # Skip build entry point
        if module_instance == build_ctx.entry:
            continue

        # Get the build context for the laid out module
        module_super_ctxs = laid_out_modules[module_super]
        if len(module_super_ctxs) > 1:
            raise errors.AtoNotImplementedError()
        module_super_ctx = module_super_ctxs[0]

        # Build up a map of UUIDs of the children of the module
        # The keys are instance UUIDs and the values are the corresponding UUIDs in the layout
        # FIXME: this currently relies on the `all_descendants` iterator returning the
        # children in the same order. This is pretty fragile and should be fixed.
        uuid_map = {}
        for inst_addr, layout_addr in zip(
            all_descendants(module_instance), all_descendants(module_super_ctx.entry)
        ):
            if not match_components(inst_addr):
                # Skip non-components
                continue

            if address.get_name(inst_addr) != address.get_name(layout_addr):
                errors.AtoError(
                    "Mismatched names in layout and instance addresses,"
                    " skipping component in auto-layout generation"
                ).log(log, logging.WARNING)
                continue

            uuid_map[generate_comp_uid(inst_addr)] = generate_comp_uid(layout_addr)

        module_map[address.get_instance_section(module_instance)] = {
            "instance_path": module_instance,
            "layout_path": str(module_super_ctx.layout_path),
            "uuid_map": uuid_map,
        }

    with open(build_ctx.output_base.with_suffix(".layouts.json"), "w", encoding="utf-8") as f:
        json.dump(module_map, f)



import pytest
from atopile.generic_methods import dfs_postorder, bfs


class Node:
    def __init__(self, name, children):
        self.name = name
        self.children = children


@pytest.fixture
def tree():
    """
    a
    |- b
    |  |- e
    |  |- f
    |
    |- c
    |  |- g
    |  |- h
    |
    |- d
    """
    a = Node("a", [])
    b = Node("b", [])
    c = Node("c", [])
    d = Node("d", [])
    e = Node("e", [])
    f = Node("f", [])
    g = Node("g", [])
    h = Node("h", [])

    a.children = [b, c, d]
    b.children = [e, f]
    c.children = [g, h]

    return a, b, c, d, e, f, g, h


def test_bfs(tree: tuple[Node]):
    a, b, c, d, e, f, g, h = tree
    assert list(bfs(lambda n: n.children, a)) == [a, b, c, d, e, f, g, h]


def test_dfs_postorder(tree: tuple[Node]):
    a, b, c, d, e, f, g, h = tree
    assert list(dfs_postorder(lambda n: n.children, a)) == [e, f, b, g, h, c, d, a]
